from typing import List, Optional, Union

from supervisely import TagMeta, TagMetaCollection, TagValueType
from supervisely.app.content import StateJson
from supervisely.app.widgets import NotificationBox, Widget
from supervisely.imaging.color import rgb2hex

VALUE_TYPE_NAME = {
    str(TagValueType.NONE): "NONE",
    str(TagValueType.ANY_STRING): "TEXT",
    str(TagValueType.ONEOF_STRING): "ONE OF",
    str(TagValueType.ANY_NUMBER): "NUMBER",
}


class TagsListSelector(Widget):
    """Widget to select one or multiple TagMetas from a provided list."""

    class Routes:
        """Callback route names used by the widget frontend to notify Python."""

        CHECKBOX_CHANGED = "checkbox_cb"

    def __init__(
        self,
        tag_metas: Optional[Union[List[TagMeta], TagMetaCollection]] = [],
        multiple: Optional[bool] = False,
        empty_notification: Optional[NotificationBox] = None,
        widget_id: Optional[str] = None,
    ):
        """:param tag_metas: List or TagMetaCollection to select from.
        :type tag_metas: Union[List[TagMeta], TagMetaCollection], optional
        :param multiple: If True, allow multiple selection.
        :type multiple: bool, optional
        :param empty_notification: Widget to show when no tags.
        :type empty_notification: NotificationBox, optional
        :param widget_id: Unique widget identifier.
        :type widget_id: str, optional
        """
        self._tag_metas = tag_metas
        self._multiple = multiple
        if empty_notification is None:
            empty_notification = NotificationBox(
                title="No tags",
                description="No tags to select.",
            )
        self.empty_notification = empty_notification
        super().__init__(widget_id=widget_id, file_path=__file__)

    def get_json_data(self):
        return {
            "tags": [
                {
                    "name": f"<b>{tag_meta.name}</b>",
                    "valueType": VALUE_TYPE_NAME[tag_meta.value_type],
                    "color": rgb2hex(tag_meta.color),
                }
                for tag_meta in self._tag_metas
            ]
        }

    def get_json_state(self):
        return {"selected": [False for _ in self._tag_metas]}

    def set(self, tags: Union[List[TagMeta], TagMetaCollection]):
        selected_tags = [cls.name for cls in self.get_selected_tags()]
        self._tag_metas = tags
        StateJson()[self.widget_id]["selected"] = [
            cls.name in selected_tags for cls in self._tag_metas
        ]
        self.update_data()
        StateJson().send_changes()

    def get_selected_tags(self):
        selected = StateJson()[self.widget_id]["selected"]
        return [cls for cls, is_selected in zip(self._tag_metas, selected) if is_selected]

    def select_all(self):
        StateJson()[self.widget_id]["selected"] = [True for _ in self._tag_metas]
        StateJson().send_changes()

    def deselect_all(self):
        # pylint: disable=no-member
        StateJson()[self.widget_id]["selected"] = [False for _ in self._tags]
        StateJson().send_changes()

    def select(self, names: List[str]):
        selected = [cls.name in names for cls in self._tag_metas]
        StateJson()[self.widget_id]["selected"] = selected
        StateJson().send_changes()

    def deselect(self, names: List[str]):
        selected = StateJson()[self.widget_id]["selected"]
        for idx, cls in enumerate(self._tag_metas):
            if cls.name in names:
                selected[idx] = False
        StateJson()[self.widget_id]["selected"] = selected
        StateJson().send_changes()

    def set_multiple(self, value: bool):
        self._multiple = value

    def get_all_tags(self):
        return self._tag_metas

    def selection_changed(self, func):
        route_path = self.get_route_path(TagsListSelector.Routes.CHECKBOX_CHANGED)
        server = self._sly_app.get_server()
        self._checkboxes_handled = True

        @server.post(route_path)
        def _click():
            selected = self.get_selected_tags()
            func(selected)

        return _click



# coding: utf-8
from __future__ import annotations

from typing import Dict, Iterator, List, Optional

from supervisely.pointcloud_annotation.pointcloud_object import PointcloudObject
from supervisely.project.project_meta import ProjectMeta
from supervisely.video_annotation.key_id_map import KeyIdMap
from supervisely.video_annotation.video_object_collection import VideoObjectCollection


class PointcloudObjectCollection(VideoObjectCollection):
    """
    Collection with :class:`~supervisely.pointcloud_annotation.pointcloud_object.PointcloudObject` instances.
    :class:`~supervisely.pointcloud_annotation.pointcloud_object_collection.PointcloudObjectCollection` object is immutable.
    """

    item_type = PointcloudObject

    def __iter__(self) -> Iterator[PointcloudObject]:
        return next(self)

    @classmethod
    def from_json(
        cls, data: List[Dict], project_meta: ProjectMeta, key_id_map: Optional[KeyIdMap] = None
    ) -> PointcloudObjectCollection:
        """
        Convert a list of json dicts to PointcloudObjectCollection. Read more about `Supervisely format <https://docs.supervisely.com/data-organization/00_ann_format_navi>`_.

        :param data: List with dicts in json format.
        :type data: List[dict]
        :param project_meta: Input project metadata.
        :type project_meta: :class:`~supervisely.project.project_meta.ProjectMeta`
        :param key_id_map: Key ID map.
        :type key_id_map: :class:`~supervisely.video_annotation.key_id_map.KeyIdMap`
        :returns: Pointcloud object collection.
        :rtype: :class:`~supervisely.pointcloud_annotation.pointcloud_object_collection.PointcloudObjectCollection`

        :Usage Example:

            .. code-block:: python

                import supervisely as sly
                from supervisely.geometry.cuboid_3d import Cuboid3d
                from supervisely.pointcloud_annotation.pointcloud_object_collection import PointcloudObjectCollection

                obj_collection_json = [
                    {"classTitle": "car", "tags": []},
                    {"classTitle": "bus", "tags": []}
                ]

                class_car = sly.ObjClass('car', Cuboid3d)
                class_bus = sly.ObjClass('bus', Cuboid3d)
                classes = sly.ObjClassCollection([class_car, class_bus])
                meta = sly.ProjectMeta(obj_classes=classes)

                pointcloud_obj_collection = sly.PointcloudObjectCollection.from_json(obj_collection_json, meta)
        """

        return super().from_json(data, project_meta, key_id_map=key_id_map)



# coding: utf-8

from typing import Optional

from supervisely.video_annotation.frame import Frame
from supervisely.video_annotation.key_id_map import KeyIdMap
from supervisely.volume_annotation.volume_figure import VolumeFigure
from supervisely.volume_annotation.volume_object_collection import VolumeObjectCollection
from supervisely.volume_annotation.constants import FIGURES, INDEX


class Slice(Frame):
    """Single slice in volume plane; holds VolumeFigures at given index. Immutable."""

    figure_type = VolumeFigure

    def __init__(self, index: int, figures: Optional[list] = None):
        """
        Same parameters as :class:`~supervisely.video_annotation.frame.Frame`.

        :param index: Slice index.
        :type index: int
        :param figures: List of VolumeFigures.
        :type figures: Optional[List[:class:`~supervisely.volume_annotation.volume_figure.VolumeFigure`]]

        :Usage Example:

            .. code-block:: python

                import supervisely as sly

                obj_class = sly.ObjClass('car', sly.Rectangle)
                volume_obj = sly.VolumeObject(obj_class)
                geometry = sly.Rectangle(0, 0, 100, 100)
                figure = sly.VolumeFigure(volume_obj, geometry, "axial", 7)
                slice_ = sly.Slice(7, figures=[figure])
        """
        super().__init__(index, figures)

    # @classmethod
    # def from_json(cls, data, objects, slices_count=None, key_id_map=None):
    #     raise NotImplementedError()
    #     _frame = super().from_json(data, objects, slices_count, key_id_map)
    #     return cls(index=_frame.index, figures=_frame.figures)

    @classmethod
    def from_json(
        cls,
        data: dict,
        objects: VolumeObjectCollection,
        plane_name: str,
        slices_count: Optional[int] = None,
        key_id_map: Optional[KeyIdMap] = None,
    ):
        """
        Deserialize a Slice object from a JSON representation.

        :param data: The JSON representation of the Slice.
        :type data: dict
        :param objects: A collection of objects in volume.
        :type objects: :class:`~supervisely.volume_annotation.volume_object_collection.VolumeObjectCollection`
        :param plane_name: The name of the plane.
        :type plane_name: str
        :param slices_count: The total number of slices in the volume, if known.
        :type slices_count: Optional[int]
        :param key_id_map: A mapping of keys to IDs used for referencing objects.
        :type key_id_map: Optional[:class:`~supervisely.video_annotation.key_id_map.KeyIdMap`]
        :returns: The deserialized Slice object.
        :rtype: :class:`~supervisely.volume_annotation.slice.Slice`
        :raises ValueError: If the slice index is negative or greater than the total number of slices.

        :Usage Example:

            .. code-block:: python

                import supervisely as sly
                slice_index = 7
                geometry = sly.Rectangle(0, 0, 100, 100)
                class_car = sly.ObjClass('car', sly.Rectangle)
                object_car = sly.VolumeObject(class_car)
                objects = sly.VolumeObjectCollection([object_car])
                figure_car = sly.VolumeFigure(object_car, geometry, sly.Plane.AXIAL, slice_index)

                slice = sly.Slice(slice_index, figures=[figure_car])
                slice_json = slice.to_json()

                new_slice = sly.Slice.from_json(slice_json, objects, sly.Plane.AXIAL)
        """

        index = data[INDEX]
        if index < 0:
            raise ValueError("Frame Index have to be >= 0")

        if slices_count is not None:
            if index > slices_count:
                raise ValueError(
                    "Item contains {} frames. Frame index is {}".format(slices_count, index)
                )

        figures = []
        for figure_json in data.get(FIGURES, []):
            figure = cls.figure_type.from_json(figure_json, objects, plane_name, index, key_id_map)
            figures.append(figure)
        return cls(index=index, figures=figures)



from supervisely.geometry.geometry import Geometry
from supervisely.geometry.constants import (
    EXTERIOR,
    INTERIOR,
    POINTS,
    LABELER_LOGIN,
    UPDATED_AT,
    CREATED_AT,
    ID,
    CLASS_ID,
)


class ClosedSurfaceMesh(Geometry):
    """3D closed surface mesh (triangular faces). Immutable."""

    @staticmethod
    def geometry_name():
        """
        Returns the name of the geometry.

        :returns: name of the geometry
        :rtype: str
        """
        return "closed_surface_mesh"

    def draw(self, bitmap, color, thickness=1, config=None):
        """
        Draws the ClosedSurfaceMesh on a bitmap.

        :param bitmap: np.ndarray
        :param color: [R, G, B]
        :param thickness: (int)
        :param config: drawing config specific to a concrete subclass, e.g. per edge colors
        :type config: dict
        """
        raise NotImplementedError('Method "draw" is unavailable for this geometry')

    def draw_contour(self, bitmap, color, thickness=1, config=None):
        """
        Draws the contour of the ClosedSurfaceMesh on a bitmap.

        :param bitmap: np.ndarray
        :param color: [R, G, B]
        :param thickness: (int)
        :param config: drawing config specific to a concrete subclass, e.g. per edge colors
        :type config: dict
        """
        raise NotImplementedError(
            'Method "draw_contour" is unavailable for this geometry'
        )

    def convert(self, new_geometry, contour_radius=0, approx_epsilon=None):
        """
        Converts the ClosedSurfaceMesh to a new geometry.

        :param new_geometry: new geometry
        :type new_geometry: :class:`~supervisely.geometry.geometry.Geometry`
        :param contour_radius: radius of the contour
        :type contour_radius: int
        :param approx_epsilon: epsilon for the approximation
        :type approx_epsilon: float
        """
        raise NotImplementedError('Method "convert" is unavailable for this geometry')

    def to_json(self):
        """
        Converts the ClosedSurfaceMesh to a JSON object.

        :returns: JSON object
        :rtype: dict
        """
        res = {}
        self._add_creation_info(res)
        return res

    @classmethod
    def from_json(cls, data):
        """
        Converts a JSON object to a ClosedSurfaceMesh.

        :param data: JSON object
        :type data: dict
        :returns: ClosedSurfaceMesh
        :rtype: :class:`~supervisely.geometry.closed_surface_mesh.ClosedSurfaceMesh`
        """
        labeler_login = data.get(LABELER_LOGIN, None)
        updated_at = data.get(UPDATED_AT, None)
        created_at = data.get(CREATED_AT, None)
        sly_id = data.get(ID, None)
        class_id = data.get(CLASS_ID, None)

        return cls(
            sly_id=sly_id,
            class_id=class_id,
            labeler_login=labeler_login,
            updated_at=updated_at,
            created_at=created_at,
        )



# vim: expandtab:ts=4:sw=4
import numpy as np
import scipy.linalg


"""
Table for the 0.95 quantile of the chi-square distribution with N degrees of
freedom (contains values for N=1, ..., 9). Taken from MATLAB/Octave's chi2inv
function and used as Mahalanobis gating threshold.
"""
chi2inv95 = {
    1: 3.8415,
    2: 5.9915,
    3: 7.8147,
    4: 9.4877,
    5: 11.070,
    6: 12.592,
    7: 14.067,
    8: 15.507,
    9: 16.919}


class KalmanFilter(object):
    """
    A simple Kalman filter for tracking bounding boxes in image space.

    The 8-dimensional state space

        x, y, w, h, vx, vy, vw, vh

    contains the bounding box center position (x, y), width w, height h,
    and their respective velocities.

    Object motion follows a constant velocity model. The bounding box location
    (x, y, w, h) is taken as direct observation of the state space (linear
    observation model).

    """

    def __init__(self):
        ndim, dt = 4, 1.

        # Create Kalman filter model matrices.
        self._motion_mat = np.eye(2 * ndim, 2 * ndim)
        for i in range(ndim):
            self._motion_mat[i, ndim + i] = dt
        self._update_mat = np.eye(ndim, 2 * ndim)

        # Motion and observation uncertainty are chosen relative to the current
        # state estimate. These weights control the amount of uncertainty in
        # the model. This is a bit hacky.
        self._std_weight_position = 1. / 20
        self._std_weight_velocity = 1. / 160

    def initiate(self, measurement):
        """Create track from unassociated measurement.

        Parameters
        ----------
        measurement : ndarray
            Bounding box coordinates (x, y, w, h) with center position (x, y),
            width w, and height h.

        Returns
        -------
        (ndarray, ndarray)
            Returns the mean vector (8 dimensional) and covariance matrix (8x8
            dimensional) of the new track. Unobserved velocities are initialized
            to 0 mean.

        """
        mean_pos = measurement
        mean_vel = np.zeros_like(mean_pos)
        mean = np.r_[mean_pos, mean_vel]

        std = [
            2 * self._std_weight_position * measurement[2],
            2 * self._std_weight_position * measurement[3],
            2 * self._std_weight_position * measurement[2],
            2 * self._std_weight_position * measurement[3],
            10 * self._std_weight_velocity * measurement[2],
            10 * self._std_weight_velocity * measurement[3],
            10 * self._std_weight_velocity * measurement[2],
            10 * self._std_weight_velocity * measurement[3]]
        covariance = np.diag(np.square(std))
        return mean, covariance

    def predict(self, mean, covariance):
        """Run Kalman filter prediction step.

        Parameters
        ----------
        mean : ndarray
            The 8 dimensional mean vector of the object state at the previous
            time step.
        covariance : ndarray
            The 8x8 dimensional covariance matrix of the object state at the
            previous time step.

        Returns
        -------
        (ndarray, ndarray)
            Returns the mean vector and covariance matrix of the predicted
            state. Unobserved velocities are initialized to 0 mean.

        """
        std_pos = [
            self._std_weight_position * mean[2],
            self._std_weight_position * mean[3],
            self._std_weight_position * mean[2],
            self._std_weight_position * mean[3]]
        std_vel = [
            self._std_weight_velocity * mean[2],
            self._std_weight_velocity * mean[3],
            self._std_weight_velocity * mean[2],
            self._std_weight_velocity * mean[3]]
        motion_cov = np.diag(np.square(np.r_[std_pos, std_vel]))

        mean = np.dot(mean, self._motion_mat.T)
        covariance = np.linalg.multi_dot((
            self._motion_mat, covariance, self._motion_mat.T)) + motion_cov

        return mean, covariance

    def project(self, mean, covariance):
        """Project state distribution to measurement space.

        Parameters
        ----------
        mean : ndarray
            The state's mean vector (8 dimensional array).
        covariance : ndarray
            The state's covariance matrix (8x8 dimensional).

        Returns
        -------
        (ndarray, ndarray)
            Returns the projected mean and covariance matrix of the given state
            estimate.

        """
        std = [
            self._std_weight_position * mean[2],
            self._std_weight_position * mean[3],
            self._std_weight_position * mean[2],
            self._std_weight_position * mean[3]]
        innovation_cov = np.diag(np.square(std))

        mean = np.dot(self._update_mat, mean)
        covariance = np.linalg.multi_dot((
            self._update_mat, covariance, self._update_mat.T))
        return mean, covariance + innovation_cov

    def multi_predict(self, mean, covariance):
        """Run Kalman filter prediction step (Vectorized version).
        Parameters
        ----------
        mean : ndarray
            The Nx8 dimensional mean matrix of the object states at the previous
            time step.
        covariance : ndarray
            The Nx8x8 dimensional covariance matrics of the object states at the
            previous time step.
        Returns
        -------
        (ndarray, ndarray)
            Returns the mean vector and covariance matrix of the predicted
            state. Unobserved velocities are initialized to 0 mean.
        """
        std_pos = [
            self._std_weight_position * mean[:, 2],
            self._std_weight_position * mean[:, 3],
            self._std_weight_position * mean[:, 2],
            self._std_weight_position * mean[:, 3]]
        std_vel = [
            self._std_weight_velocity * mean[:, 2],
            self._std_weight_velocity * mean[:, 3],
            self._std_weight_velocity * mean[:, 2],
            self._std_weight_velocity * mean[:, 3]]
        sqr = np.square(np.r_[std_pos, std_vel]).T

        motion_cov = []
        for i in range(len(mean)):
            motion_cov.append(np.diag(sqr[i]))
        motion_cov = np.asarray(motion_cov)

        mean = np.dot(mean, self._motion_mat.T)
        left = np.dot(self._motion_mat, covariance).transpose((1, 0, 2))
        covariance = np.dot(left, self._motion_mat.T) + motion_cov

        return mean, covariance

    def update(self, mean, covariance, measurement):
        """Run Kalman filter correction step.

        Parameters
        ----------
        mean : ndarray
            The predicted state's mean vector (8 dimensional).
        covariance : ndarray
            The state's covariance matrix (8x8 dimensional).
        measurement : ndarray
            The 4 dimensional measurement vector (x, y, w, h), where (x, y)
            is the center position, w the width, and h the height of the
            bounding box.

        Returns
        -------
        (ndarray, ndarray)
            Returns the measurement-corrected state distribution.

        """
        projected_mean, projected_cov = self.project(mean, covariance)

        chol_factor, lower = scipy.linalg.cho_factor(
            projected_cov, lower=True, check_finite=False)
        kalman_gain = scipy.linalg.cho_solve(
            (chol_factor, lower), np.dot(covariance, self._update_mat.T).T,
            check_finite=False).T
        innovation = measurement - projected_mean

        new_mean = mean + np.dot(innovation, kalman_gain.T)
        new_covariance = covariance - np.linalg.multi_dot((
            kalman_gain, projected_cov, kalman_gain.T))
        return new_mean, new_covariance

    def gating_distance(self, mean, covariance, measurements,
                        only_position=False, metric='maha'):
        """Compute gating distance between state distribution and measurements.
        A suitable distance threshold can be obtained from `chi2inv95`. If
        `only_position` is False, the chi-square distribution has 4 degrees of
        freedom, otherwise 2.
        Parameters
        ----------
        mean : ndarray
            Mean vector over the state distribution (8 dimensional).
        covariance : ndarray
            Covariance of the state distribution (8x8 dimensional).
        measurements : ndarray
            An Nx4 dimensional matrix of N measurements, each in
            format (x, y, a, h) where (x, y) is the bounding box center
            position, a the aspect ratio, and h the height.
        only_position : Optional[bool]
            If True, distance computation is done with respect to the bounding
            box center position only.
        Returns
        -------
        ndarray
            Returns an array of length N, where the i-th element contains the
            squared Mahalanobis distance between (mean, covariance) and
            `measurements[i]`.
        """
        mean, covariance = self.project(mean, covariance)
        if only_position:
            mean, covariance = mean[:2], covariance[:2, :2]
            measurements = measurements[:, :2]

        d = measurements - mean
        if metric == 'gaussian':
            return np.sum(d * d, axis=1)
        elif metric == 'maha':
            cholesky_factor = np.linalg.cholesky(covariance)
            z = scipy.linalg.solve_triangular(
                cholesky_factor, d.T, lower=True, check_finite=False,
                overwrite_b=True)
            squared_maha = np.sum(z * z, axis=0)
            return squared_maha
        else:
            raise ValueError('invalid distance metric')


import numpy as np
from typing import Dict, Optional, Any, Tuple, Mapping

import supervisely as sly
from supervisely import logger
from supervisely.nn import TaskType

from .metrics import SegmentationMetrics, DetectionMetrics

class LiveEvaluator:
    """
    App-agnostic evaluator for Live Training.

    Stores predictions (sly objects) by image_id, converts them to task-specific format,
    computes metrics when ground truth arrives, and tracks values using EMA.
    """

    def __init__(
        self,
        task_type: TaskType,
        class2idx: Mapping[str, int],
        ema_alpha: float = 0.1,
        ignore_index: int = 255,
        score_thr: float = None,
    ):
        """
        :param task_type: SEMANTIC_SEGMENTATION or OBJECT_DETECTION.
        :type task_type: TaskType
        :param class2idx: Class name to index.
        :type class2idx: Mapping[str, int]
        :param ema_alpha: EMA smoothing.
        :type ema_alpha: float
        :param ignore_index: Ignored class index.
        :type ignore_index: int
        :param score_thr: Detection score threshold. :param score_thr: Detection score threshold.
        :type score_thr: float
        """
        if not 0 < ema_alpha <= 1:
            raise ValueError(f"ema_alpha must be in (0, 1], got {ema_alpha}")

        self.task_type: TaskType = task_type
        self.class2idx: Mapping[str, int] = class2idx
        self.ema_alpha: float = ema_alpha
        self.ignore_index: int = ignore_index
        self.score_thr: float = score_thr

        self._predictions: Dict[Any, Tuple[list, Tuple[int, int]]] = {}

        self._sample_count = 0
        self.ema_value: Optional[float] = None

    def store_prediction(self, image_id: Any, objects: list, image_shape: Tuple[int, int]):
        if image_id in self._predictions:
            logger.warning(f"Prediction for image_id={image_id} already exists, overwriting")
        self._predictions[image_id] = (objects, image_shape)

    def evaluate(
        self, image_id: Any, ground_truth_annotation: sly.Annotation
    ) -> Optional[Dict[str, float]]:
        if image_id not in self._predictions:
            logger.warning(f"No prediction stored for image_id={image_id}. Skipping evaluation")
            return None

        objects, image_shape = self._predictions.pop(image_id)

        if self.task_type == TaskType.SEMANTIC_SEGMENTATION:
            metrics = self._evaluate_segmentation(objects, ground_truth_annotation, image_shape)
        elif self.task_type == TaskType.OBJECT_DETECTION:
            metrics = self._evaluate_detection(objects, ground_truth_annotation, image_shape)
        else:
            raise ValueError(f"Unsupported task_type: {self.task_type}")

        metric_value = metrics[self.metric_name]
        if metric_value is not None:
            self._update_ema(metric_value)

        return {'metric_value': metric_value, 'ema_value': self.ema_value}

    def _evaluate_segmentation(self, objects: list, gt_annotation: sly.Annotation, image_shape: Tuple[int,int]) -> float:
        pred_mask = self._pred_objects_to_mask(objects, image_shape)
        gt_mask = self._gt_annotation_to_mask(gt_annotation, image_shape)
        metrics_calc = SegmentationMetrics(num_classes=len(self.class2idx), ignore_index=self.ignore_index)
        metrics = metrics_calc.all_metrics(pred_mask, gt_mask)
        return metrics

    def _evaluate_detection(self, objects: list, gt_annotation: sly.Annotation, image_shape: Tuple[int,int]) -> float:
        if self.score_thr is not None:
            objects = [obj for obj in objects if obj.get('meta', {}).get('confidence') >= self.score_thr]
        pred_boxes, pred_labels = self._pred_objects_to_bboxes(objects)
        gt_boxes, gt_labels = self._gt_annotation_to_bboxes(gt_annotation)
        metrics_calc = DetectionMetrics()
        metrics = metrics_calc.all_metrics(pred_boxes, pred_labels, gt_boxes, gt_labels)
        return metrics

    def _pred_objects_to_mask(self, objects: list, image_shape: Tuple[int,int]) -> np.ndarray:
        """Convert predicted objects to segmentation mask."""
        height, width = image_shape
        mask = np.zeros((height, width), dtype=np.int32)

        project_meta = self._get_project_meta_stub()
        for obj in objects:
            obj_class_title = obj['classTitle']
            if obj_class_title is None:
                continue

            class_idx = self.class2idx.get(obj_class_title)
            if class_idx is None:
                continue

            try:
                label = sly.Label.from_json(obj, project_meta)
                label.geometry.draw(mask, color=class_idx)
            except Exception as e:
                logger.warning(f"Failed to parse object geometry: {e}")
        return mask

    def _gt_annotation_to_mask(self, annotation: sly.Annotation, image_shape: Tuple[int,int]) -> np.ndarray:
        """Convert ground truth annotation to segmentation mask."""
        height, width = image_shape
        mapping = {label.obj_class: label.obj_class for label in annotation.labels}
        ann_nonoverlap = annotation.to_nonoverlapping_masks(mapping)

        mask = np.zeros((height, width), dtype=np.int32)
        for label in ann_nonoverlap.labels:
            class_name = label.obj_class.name
            class_idx = self.class2idx.get(class_name)
            if class_idx is not None:
                try:
                    label.geometry.draw(mask, color=class_idx)
                except Exception as e:
                    logger.warning(f"Failed to draw label geometry: {e}")
        return mask

    def _pred_objects_to_bboxes(self, objects: list):
        bboxes = []
        labels = []

        project_meta = self._get_project_meta_stub()

        for obj in objects:
            class_title = obj.get('classTitle')
            if class_title is None:
                continue

            class_idx = self.class2idx.get(class_title)
            if class_idx is None:
                continue

            try:
                label = sly.Label.from_json(obj, project_meta)
                geom = label.geometry
                if not isinstance(geom, sly.Rectangle):
                    continue  

                bbox = [geom.left, geom.top, geom.right, geom.bottom]
                bboxes.append(bbox)
                labels.append(class_idx)
            except Exception as e:
                logger.warning(f"Failed to parse prediction bbox: {e}")

        return (np.array(bboxes, dtype=np.float32), np.array(labels, dtype=np.int64)) 

    def _gt_annotation_to_bboxes(self, annotation: sly.Annotation):
        bboxes = []
        labels = []

        for label in annotation.labels:
            class_idx = self.class2idx.get(label.obj_class.name)
            if class_idx is None:
                continue
            geom = label.geometry
            if not isinstance(geom, sly.Rectangle):
                continue

            bboxes.append([geom.left, geom.top, geom.right, geom.bottom])
            labels.append(class_idx)

        return (np.array(bboxes, dtype=np.float32), np.array(labels, dtype=np.int64))

    def _get_project_meta_stub(self) -> sly.ProjectMeta:
        obj_classes = [sly.ObjClass(class_name, sly.AnyGeometry) for class_name in self.class2idx.keys()]
        return sly.ProjectMeta(obj_classes=sly.ObjClassCollection(obj_classes))

    def _update_ema(self, new_value: float):
        self._sample_count += 1
        alpha = min(self.ema_alpha * 2, 1.0) if self._sample_count <= 2 else self.ema_alpha
        if self.ema_value is None:
            self.ema_value = new_value
        else:
            self.ema_value = alpha * new_value + (1 - alpha) * self.ema_value

    def reset(self):
        self._predictions.clear()
        self.ema_value = None

    def state_dict(self) -> Dict:
        return {
            'task_type': self.task_type,
            'ema_alpha': self.ema_alpha,
            'ema_value': self.ema_value,
            'ignore_index': self.ignore_index
        }

    def load_state_dict(self, state: Dict):
        self.ema_value = state.get('ema_value')

    @property
    def metric_name(self) -> str:
        """Primary metric name for this task type"""
        if self.task_type == TaskType.SEMANTIC_SEGMENTATION:
            return 'mean_boundary_iou'
        elif self.task_type == TaskType.OBJECT_DETECTION:
            return 'assistance_score'
        else:
            raise ValueError(f"Unsupported task_type: {self.task_type}")



from __future__ import annotations
from supervisely.app.jinja2 import create_env
from supervisely.app.content import DataJson, StateJson
from supervisely.app.widgets import Widget, ConditionalWidget, ConditionalItem
from typing import List, Dict, Optional

try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal


class RadioGroup(ConditionalWidget):
    """Radio-button group widget that selects one of provided items (optionally with per-item content)."""

    class Routes:
        """Callback route names used by the widget frontend to notify Python."""

        VALUE_CHANGED = "value_changed"
    
    class Item(ConditionalItem):
        """Item type alias for radio group options."""

        pass

    def __init__(
        self,
        items: List[RadioGroup.Item],
        size: Literal["large", "small", "mini"] = None,
        direction: Literal["vertical", "horizontal"] = "horizontal",
        gap: int = 10,
        widget_id: str = None,
    ) -> RadioGroup:
        """:param items: List of RadioGroup.Item (value, label, optional content).
        :type items: List[RadioGroup.Item]
        :param size: Button size: "large", "small", or "mini".
        :type size: Literal["large", "small", "mini"], optional
        :param direction: "vertical" or "horizontal" layout.
        :type direction: Literal["vertical", "horizontal"]
        :param gap: Gap between buttons in pixels.
        :type gap: int
        :param widget_id: Unique widget identifier.
        :type widget_id: str, optional
        """
        self._changes_handled = False
        self._size = size
        self._gap = gap
        self._value = None

        self._flex_direction = "row"
        if direction == "vertical":
            self._flex_direction = "column"

        super().__init__(items=items, widget_id=widget_id, file_path=__file__)

    def _get_first_value(self) -> RadioGroup.Item:
        if self._items is not None and len(self._items) > 0:
            return self._items[0]
        return None

    def get_json_data(self) -> Dict:
        res = {
            "items": None,
        }
        if self._items is not None:
            res["items"] = [item.to_json() for item in self._items]
        if self._size is not None:
            res["size"] = self._size
        return res

    def get_json_state(self) -> Dict:
        value = self._value
        if value is None:
            first_item = self._get_first_value()
            if first_item is not None:
                value = first_item.value
        return {"value": value}

    def get_value(self):
        return StateJson()[self.widget_id]["value"]

    def value_changed(self, func):
        route_path = self.get_route_path(RadioGroup.Routes.VALUE_CHANGED)
        server = self._sly_app.get_server()
        self._changes_handled = True

        @server.post(route_path)
        def _click():
            res = self.get_value()
            self._value = res
            func(res)

        return _click

    def set(self, items: List[RadioGroup.Item]):
        self._items = items
        self.update_data()
        self.update_state()
        DataJson().send_changes()
        StateJson().send_changes()

    def set_value(self, value):
        self._value = value
        self.update_state()
        StateJson().send_changes()



# coding: utf-8

from supervisely.geometry.geometry import Geometry
from supervisely.geometry.constants import ANY_SHAPE


class AnyGeometry(Geometry):
    """Placeholder geometry that accepts any shape type. Used for labels with flexible geometry. Immutable."""
    @staticmethod
    def geometry_name():
        """
        Geometry name.

        :returns: Geometry name
        :rtype: str
        """
        return ANY_SHAPE



# coding: utf-8
"""List user roles available in a Supervisely instance."""

# docs
from __future__ import annotations

from enum import IntEnum
from typing import Dict, List, NamedTuple, Optional

from supervisely.api.module_api import ApiField, ModuleApiBase


class RoleInfo(NamedTuple):
    """NamedTuple describing a role entry returned by the API."""

    id: int
    role: str
    created_at: str
    updated_at: str


class RoleApi(ModuleApiBase):
    """API for working with roles."""

    def __init__(self, api):
        """
        :param api: :class:`~supervisely.api.api.Api` object to use for API connection.
        :type api: :class:`~supervisely.api.api.Api`

        :Usage Example:

            .. code-block:: python

                import supervisely as sly
                api = sly.Api.from_env()
                roles = api.role.get_list()
        """
        super().__init__(api)

    class DefaultRole(IntEnum):
        """Built-in role IDs used by the platform (admin/developer/annotator/viewer)."""

        ADMIN = 1
        """"""
        DEVELOPER = 2
        """"""
        ANNOTATOR = 3
        """"""
        VIEWER = 4
        """"""

    @staticmethod
    def info_sequence():
        """
        Sequence of fields that are returned by the API to represent RoleInfo.

        :Usage Example:

            .. code-block:: python

                RoleInfo(
                    id=71,
                    role='manager',
                    created_at='2019-12-10T14:31:41.878Z',
                    updated_at='2019-12-10T14:31:41.878Z'
                )
        """
        return [ApiField.ID, ApiField.ROLE, ApiField.CREATED_AT, ApiField.UPDATED_AT]

    @staticmethod
    def info_tuple_name():
        """
        Name of the tuple that represents RoleInfo.
        """
        return "RoleInfo"

    def get_list(self, filters: Optional[List[Dict[str, str]]] = None) -> List[RoleInfo]:
        """
        List of all roles that are available on the Supervisely instance.

        :param filters: List of params to sort output Roles.
        :type filters: List[Dict[str, str]]
        :returns: List of all roles with information.
        :rtype: List[:class:`~supervisely.api.role_api.RoleInfo`]

        :Usage Example:

            .. code-block:: python

                import os
                from dotenv import load_dotenv

                import supervisely as sly

                # Load secrets and create API object from .env file (recommended)
                # Learn more here: https://developer.supervisely.com/getting-started/basics-of-authentication
                if sly.is_development():
                    load_dotenv(os.path.expanduser("~/supervisely.env"))

                api = sly.Api.from_env()

                roles = api.role.get_list()
        """
        return self.get_list_all_pages("roles.list", {ApiField.FILTER: filters or []})

    def _convert_json_info(self, info: dict, skip_missing=True):
        """ """
        res = super()._convert_json_info(info, skip_missing=skip_missing)
        return RoleInfo(**res._asdict())



# coding: utf-8
"""collection with :class:`~supervisely.annotation.obj_class.ObjClass` instances"""

# docs
from __future__ import annotations

from collections import defaultdict
from typing import Dict, Iterator, List, Optional

from supervisely import logger
from supervisely.annotation.obj_class import ObjClass
from supervisely.annotation.renamer import Renamer
from supervisely.collection.key_indexed_collection import KeyIndexedCollection
from supervisely.geometry.graph import GraphNodes
from supervisely.imaging.color import hex2rgb, rgb2hex
from supervisely.io.json import JsonSerializable


class ObjClassCollection(KeyIndexedCollection, JsonSerializable):
    """Collection with :class:`~supervisely.annotation.obj_class.ObjClass` instances. :class:`~supervisely.annotation.obj_class_collection.ObjClassCollection` object is immutable."""

    item_type = ObjClass

    def __init__(self, items: Optional[List[ObjClass]] = None):
        """
        :param items: List of :class:`~supervisely.annotation.obj_class.ObjClass` instances.
        :type items: list, optional
        :raises :class:`~supervisely.collection.key_indexed_collection.DuplicateKeyError`: if instance with given name already exist

        :Usage Example:

            .. code-block:: python

                import supervisely as sly

                # Create ObjClass (see class ObjClass for more information)
                class_lemon = sly.ObjClass('lemon', sly.Rectangle)
                class_kiwi = sly.ObjClass('kiwi', sly.Bitmap)

                class_arr = [class_lemon, class_kiwi]

                # Create ObjClassCollection from ObjClasses
                classes = sly.ObjClassCollection(class_arr)

                # Add items to ObjClassCollection
                class_potato = sly.ObjClass('potato', sly.Bitmap)

                # Remember that ObjClassCollection is immutable, and we need to assign new instance of ObjClassCollection to a new variable
                classes = classes.add(class_potato)

                # You can also add multiple items to collection
                class_cabbage = sly.ObjClass('cabbage', sly.Rectangle)
                class_carrot = sly.ObjClass('carrot', sly.Bitmap)
                class_turnip = sly.ObjClass('turnip', sly.Polygon)

                classes = classes.add_items([class_cabbage, class_carrot, class_turnip])

                # Has key, checks if given key exist in collection
                classes.has_key("cabbage")
                # Output: True

                # Intersection, finds intersection of given list of instances with collection items
                class_dog = sly.ObjClass('dog', sly.Rectangle)
                class_cat = sly.ObjClass('cat', sly.Rectangle)
                class_turtle = sly.ObjClass('turtle', sly.Rectangle)

                classes_animals = sly.ObjClassCollection([class_dog, class_cat, class_turtle])

                classes_intersections = classes.intersection(classes_animals)
                print(classes_intersections.to_json())
                # Output: []

                # Let's add the potato ObjClass from another collection and compare them again
                classes_animals = classes_animals.add(class_potato)

                classes_intersections = classes.intersection(classes_animals)
                print(classes_intersections.to_json())
                # Output: [
                #     {
                #         "title":"potato",
                #         "shape":"bitmap",
                #         "color":"#8A570F",
                #         "geometry_config":{},
                #         "hotkey":""
                #     }
                # ]

                # Difference, finds difference between collection and given list of ObjClass or ObjClassCollection
                class_car = sly.ObjClass('car', sly.Rectangle)
                class_bicycle = sly.ObjClass('bicycle', sly.Rectangle)

                classes_vehicles = sly.ObjClassCollection([class_car, class_bicycle])

                class_pedestrian = sly.ObjClass('pedestrian', sly.Rectangle)
                class_road = sly.ObjClass('road', sly.Rectangle)

                difference = classes_vehicles.difference([class_pedestrian, class_road])
                print(difference.to_json())
                # Output: [
                #     {
                #         "title":"car",
                #         "shape":"rectangle",
                #         "color":"#8A0F3B",
                #         "geometry_config":{},
                #         "hotkey":""
                #     },
                #     {
                #         "title":"bicycle",
                #         "shape":"rectangle",
                #         "color":"#0F8A1F",
                #         "geometry_config":{},
                #         "hotkey":""
                #     }
                # ]

                # Merge, merges collection and given list of ObjClasses
                c_1 = sly.ObjClassCollection([class_car, class_bicycle])
                c_2 = sly.ObjClassCollection([class_pedestrian, class_road])

                c_3 = c_1.merge(c_2)
                print(c_3.to_json())
                # Output: [
                #     {
                #         "title":"pedestrian",
                #         "shape":"rectangle",
                #         "color":"#8A0F27",
                #         "geometry_config":{},
                #         "hotkey":""
                #     },
                #     {
                #         "title":"road",
                #         "shape":"rectangle",
                #         "color":"#8A620F",
                #         "geometry_config":{},
                #         "hotkey":""
                #     },
                #     {
                #         "title":"car",
                #         "shape":"rectangle",
                #         "color":"#8A0F3B",
                #         "geometry_config":{},
                #         "hotkey":""
                #     },
                #     {
                #         "title":"bicycle",
                #         "shape":"rectangle",
                #         "color":"#0F8A1F",
                #         "geometry_config":{},
                #         "hotkey":""
                #     }
                # ]

                # Merge will raise ValueError if item name from given list is in collection but items in both are different
                class_bicycle_1 = sly.ObjClass('bicycle', sly.Rectangle)
                class_bicycle_2 = sly.ObjClass('bicycle', sly.Bitmap)

                classes_1 = sly.ObjClassCollection([class_bicycle_1])
                classes_2 = sly.ObjClassCollection([class_bicycle_2])

                test_merge = classes_1.merge(classes_2)
                # Output: ValueError: Error during merge for key 'bicycle': values are different

                # Let's try to create now a collection where ObjClasses have identical names
                class_cow = sly.ObjClass('cow', sly.Rectangle)
                class_chicken = sly.ObjClass('cow', sly.Rectangle)

                test_classes = sly.ObjClassCollection([class_cow, class_chicken])
                # Output: DuplicateKeyError: "Key 'cow' already exists"
        """
        super().__init__(items)

    def to_json(self) -> List[Dict]:
        """
        Convert the ObjClassCollection to a list of json dicts. Read more about `Supervisely format <https://docs.supervisely.com/data-organization/00_ann_format_navi>`_.

        :returns: List of dicts in json format
        :rtype: :class:`List[dict]`

        :Usage Example:

            .. code-block:: python

                import supervisely as sly

                class_lemon = sly.ObjClass('lemon', sly.Rectangle)
                class_kiwi = sly.ObjClass('kiwi', sly.Bitmap)

                # Add ObjClasses to ObjClassCollection
                classes = sly.ObjClassCollection([class_lemon, class_kiwi])

                classes_json = classes.to_json()
                print(classes_json)
                # Output: [
                #      {
                #           "title": "lemon",
                #           "shape": "rectangle",
                #           "color": "#300F8A",
                #           "geometry_config": {},
                #           "hotkey": ""
                #      },
                #      {
                #           "title": "kiwi",
                #           "shape": "bitmap",
                #           "color": "#7C0F8A",
                #           "geometry_config": {},
                #           "hotkey": ""
                #      }
                # ]
        """
        return [obj_class.to_json() for obj_class in self]

    @classmethod
    def from_json(cls, data: List[Dict]) -> ObjClassCollection:
        """
        Convert a list with dicts in json format to ObjClassCollection. Read more about `Supervisely format <https://docs.supervisely.com/data-organization/00_ann_format_navi>`_.

        :param data: List with dicts in json format.
        :type data: List[dict]
        :returns: ObjClassCollection object
        :rtype: :class:`~supervisely.annotation.obj_class_collection.ObjClassCollection`

        :Usage Example:

            .. code-block:: python

                import supervisely as sly

                data = [
                    {
                        "title": "lemon",
                        "shape": "rectangle",
                        "color": "#300F8A",
                        "hotkey": ""
                    },
                    {
                        "title": "kiwi",
                        "shape": "bitmap",
                        "color": "#7C0F8A",
                        "hotkey": ""
                    }
                ]

                classes = sly.ObjClassCollection.from_json(data)
        """
        obj_classes = [ObjClass.from_json(obj_class_json) for obj_class_json in data]
        return cls(obj_classes)

    def validate_classes_colors(self, logger: Optional[logger] = None) -> str or None:
        """
        Checks for unique colors in the ObjClassCollection.

        :param logger: Input logger.
        :type logger: logger, optional
        :returns: Notification if there are objects with the same colors, otherwise None
        :rtype: str or None

        :Usage Example:

            .. code-block:: python

                import supervisely as sly

                # Let's create 2 ObjClasses with the same color
                class_lemon = sly.ObjClass('lemon', sly.Rectangle, [0, 0, 0])
                class_kiwi = sly.ObjClass('kiwi', sly.Bitmap, [0, 0, 0])

                # Add them to ObjClassCollection
                classes = sly.ObjClassCollection([class_lemon, class_kiwi])

                print(classes.validate_classes_colors())
                # Output: Classes ['lemon', 'kiwi'] have the same RGB color = [0, 0, 0]

                # Now let's change colors of our ObjClasses
                class_lemon = sly.ObjClass('lemon', sly.Rectangle, [255, 0, 0])
                class_kiwi = sly.ObjClass('kiwi', sly.Bitmap, [0, 0, 255])

                classes = sly.ObjClassCollection([class_lemon, class_kiwi])

                print(classes.validate_classes_colors())
                # Output: None
        """
        color_names = defaultdict(list)
        for obj_class in self:
            hex = rgb2hex(obj_class.color)
            color_names[hex].append(obj_class.name)

        class_colors_notify = None
        for hex_color, class_names in color_names.items():
            if len(class_names) > 1:
                warn_str = "Classes {!r} have the same RGB color = {!r}".format(
                    class_names, hex2rgb(hex_color)
                )
                if logger is not None:
                    logger.warn(warn_str)
                if class_colors_notify is None:
                    class_colors_notify = ""
                class_colors_notify += warn_str + "\n\n"
        return class_colors_notify

    def __iter__(self) -> Iterator[ObjClass]:
        return next(self)

    def refresh_ids_from(self, classes: ObjClassCollection) -> None:
        """Updates ids of classes in the collection from given classes.

        :param classes: Collection with classes to update ids from.
        :type classes: :class:`~supervisely.annotation.obj_class_collection.ObjClassCollection`
        :returns: None
        :rtype: None
        """
        for new_class in classes:
            my_class = self.get(new_class.name)
            if my_class is None:
                continue
            my_class._set_id(new_class.sly_id)


def make_renamed_classes(
    src_obj_classes: ObjClassCollection,
    renamer: Renamer,
    skip_missing: Optional[bool] = False,
) -> ObjClassCollection:
    """Returns a new ObjClassCollection with renamed classes.

    :param src_obj_classes: ObjClassCollection to rename.
    :type src_obj_classes: Input ObjClassCollection
    :param renamer: :class:`~supervisely.annotation.renamer.Renamer` object, which will handle renaming process.
    :type renamer: :class:`~supervisely.annotation.renamer.Renamer`
    :param skip_missing: If True, missing classes will be skipped, otherwise KeyError will be raised.
    :type skip_missing: Optional[bool]
    :returns: New ObjClassCollection with renamed classes.
    :rtype: :class:`~supervisely.annotation.obj_class_collection.ObjClassCollection`
    :raises KeyError: If skip_missing is False and some classes could not be renamed.
    """
    renamed_classes = []
    for src_cls in src_obj_classes:
        renamed_name = renamer.rename(src_cls.name)
        if renamed_name is not None:
            renamed_classes.append(src_cls.clone(name=renamed_name))
        elif not skip_missing:
            raise KeyError(
                "Object class name {} could not be mapped to a destination name.".format(
                    src_cls.name
                )
            )
    return ObjClassCollection(items=renamed_classes)



# coding: utf-8
from supervisely.annotation.obj_class import ObjClass
from supervisely.annotation.obj_class_collection import ObjClassCollection
from supervisely.annotation.renamer import Renamer


class ObjClassMapper:
    """Strategy interface for mapping a source :class:`~supervisely.annotation.obj_class.ObjClass` to a destination one."""

    def map(self, src: ObjClass) -> ObjClass:
        """
        """
        raise NotImplementedError()


class RenamingObjClassMapper(ObjClassMapper):
    """Map object classes by renaming their names using a :class:`~supervisely.annotation.renamer.Renamer`."""

    def __init__(self, dest_obj_classes: ObjClassCollection, renamer: Renamer):
        """
        :param dest_obj_classes: :class:`~supervisely.annotation.obj_class_collection.ObjClassCollection` object to map to.
        :type dest_obj_classes: :class:`~supervisely.annotation.obj_class_collection.ObjClassCollection`
        :param renamer: :class:`~supervisely.annotation.renamer.Renamer` object to use for renaming.
        :type renamer: :class:`~supervisely.annotation.renamer.Renamer`
        """
        self._dest_obj_classes = dest_obj_classes
        self._renamer = renamer

    def map(self, src: ObjClass) -> ObjClass:
        """
        The function map rename ObjClass in given collection
        :returns: ObjClass object
        :rtype: :class:`~supervisely.annotation.obj_class.ObjClass`
        """
        dest_name = self._renamer.rename(src.name)
        return self._dest_obj_classes.get(dest_name, None) if (dest_name is not None) else None



# coding: utf-8

from supervisely.geometry.cuboid_3d import Vector3d
from supervisely.geometry.geometry import Geometry
from supervisely.geometry.constants import LABELER_LOGIN, UPDATED_AT, CREATED_AT, ID, CLASS_ID


class Point3d(Geometry):
    """3D point (x, y, z) in volume or point cloud space. Immutable."""

    @staticmethod
    def geometry_name():
        """
        Returns the name of the geometry.

        :returns: name of the geometry
        :rtype: str
        """
        return 'point_3d'

    def __init__(self, point: Vector3d,
                 sly_id=None, class_id=None, labeler_login=None, updated_at=None, created_at=None):
        """
        Point3d is a geometry for a single :class:`~supervisely.annotation.label.Label`. :class:`~supervisely.geometry.point_3d.Point3d` object is immutable.

        :param point: Vector3d.
        :type point: :class:`~supervisely.geometry.cuboid_3d.Vector3d`
        :param sly_id: Point3d ID in Supervisely server.
        :type sly_id: int, optional
        :param class_id: ID of ObjClass to which Point3d belongs.
        :type class_id: int, optional
        :param labeler_login: Login of the user who created Point3d.
        :type labeler_login: str, optional
        :param updated_at: Date and Time when Point3d was modified last. Date Format: Year:Month:Day:Hour:Minute:Seconds. Example: '2021-01-22T19:37:50.158Z'.
        :type updated_at: str, optional
        :param created_at: Date and Time when Point3d was created. Date Format is the same as in "updated_at" parameter.
        :type created_at: str, optional
        """
        super().__init__(sly_id=sly_id, class_id=class_id, labeler_login=labeler_login, updated_at=updated_at, created_at=created_at)

        if type(point) is not Vector3d:
            raise TypeError("\"position\" param has to be of type {!r}".format(type(Vector3d)))

        self._point = point

    @property
    def x(self):
        """
        Copy of the x coordinate of the Point3d.

        :returns: x coordinate of the :class:`~supervisely.geometry.point_3d.Point3d`
        :rtype: int
        """
        return self._point._x

    @property
    def y(self):
        """
        Copy of the y coordinate of the Point3d.

        :returns: y coordinate of the :class:`~supervisely.geometry.point_3d.Point3d`
        :rtype: int
        """
        return self._point._y

    @property
    def z(self):
        """
        Copy of the z coordinate of the Point3d.

        :returns: z coordinate of the :class:`~supervisely.geometry.point_3d.Point3d`
        :rtype: int
        """
        return self._point._z

    def to_json(self):
        """
        Converts the Point3d to a JSON object.

        :returns: JSON object
        :rtype: dict
        :returns: Point3d in json format
        """
        res = self._point.to_json()
        self._add_creation_info(res)
        return res

    @classmethod
    def from_json(cls, data):
        """
        Converts a JSON object to a Point3d.

        :param data: JSON object
        :type data: dict
        :returns: Point3d
        :rtype: :class:`~supervisely.geometry.point_3d.Point3d`
        :returns: Point3d from json.
        :rtype: :class:`~supervisely.geometry.point_3d.Point3d`
        """
        point = Vector3d.from_json(data)

        labeler_login = data.get(LABELER_LOGIN, None)
        updated_at = data.get(UPDATED_AT, None)
        created_at = data.get(CREATED_AT, None)
        sly_id = data.get(ID, None)
        class_id = data.get(CLASS_ID, None)
        return cls(point, sly_id=sly_id, class_id=class_id, labeler_login=labeler_login, updated_at=updated_at, created_at=created_at)



from typing import List

from supervisely.nn.benchmark.base_visualizer import BaseVisMetrics
from supervisely.nn.benchmark.semantic_segmentation.evaluator import (
    SemanticSegmentationEvalResult,
)
from supervisely.nn.benchmark.visualization.widgets import ChartWidget, MarkdownWidget


class FrequentlyConfused(BaseVisMetrics):
    """Frequently confused class pairs chart for segmentation comparison."""


    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.eval_results: List[SemanticSegmentationEvalResult]
        self.clickable = True
        self._keypair_sep = "-"

    @property
    def md(self) -> MarkdownWidget:
        if self.is_empty:
            text = self.vis_texts.markdown_frequently_confused_empty
        else:
            text = self.vis_texts.markdown_frequently_confused
        return MarkdownWidget("frequently_confused", "Frequently Confused Classes", text=text)

    @property
    def chart(self) -> ChartWidget:
        return ChartWidget("frequently_confused", self.get_figure())

    @property
    def is_empty(self) -> bool:
        return all(len(e.mp.frequently_confused[0]) == 0 for e in self.eval_results)

    def get_figure(self):
        import numpy as np
        import plotly.graph_objects as go  # pylint: disable=import-error

        fig = go.Figure()

        classes = self.eval_results[0].classes_whitelist

        model_cnt = len(self.eval_results)
        all_models_cmat = np.zeros((model_cnt, len(classes), len(classes)))
        for model_idx, eval_result in enumerate(self.eval_results):
            cmat, _ = eval_result.mp.confusion_matrix
            all_models_cmat[model_idx] = cmat[::-1].copy()

        sum_cmat = all_models_cmat.sum(axis=0)
        np.fill_diagonal(sum_cmat, 0)
        sum_cmat_flat = sum_cmat.flatten()
        sorted_indices = np.argsort(sum_cmat_flat)[::-1]
        n_pairs = min(10, len(classes) * (len(classes) - 1))
        sorted_indices = sorted_indices[:n_pairs]
        rows = sorted_indices // len(classes)
        cols = sorted_indices % len(classes)
        labels = [f"{classes[rows[i]]}-{classes[cols[i]]}" for i in range(n_pairs)]
        for model_idx, eval_result in enumerate(self.eval_results):
            cmat = all_models_cmat[model_idx]
            probs = cmat[rows, cols] 
            probs = probs * 100
            fig.add_trace(
                go.Bar(
                    name=eval_result.name,
                    x=labels,
                    y=probs,
                    hovertemplate="%{x}: %{y:.2f}%<extra></extra>",
                    marker=dict(color=eval_result.color, line=dict(width=0.7)),
                )
            )

        return fig



import copy
from typing import Optional
import uuid

from supervisely.annotation.annotation import Annotation
from supervisely.app import DataJson
from supervisely.app.widgets import Widget
from supervisely.app.widgets import GridGallery


class LabeledImage(GridGallery):
    """Single-image gallery view with annotation overlay, opacity slider, and zoom for detailed inspection."""

    def __init__(
        self,
        annotations_opacity: float = 0.5,
        show_opacity_slider: bool = True,
        enable_zoom: bool = False,
        resize_on_zoom: bool = False,
        fill_rectangle: bool = True,
        border_width: int = 3,
        view_height: Optional[int] = None,
        widget_id: str = None,
        empty_message: str = "No image was provided",
    ):
        """
        :param annotations_opacity: Opacity of annotation overlays (0–1).
        :type annotations_opacity: float
        :param show_opacity_slider: If True, show opacity slider.
        :type show_opacity_slider: bool
        :param enable_zoom: If True, enable zoom.
        :type enable_zoom: bool
        :param resize_on_zoom: If True, resize on zoom.
        :type resize_on_zoom: bool
        :param fill_rectangle: If True, fill rectangles.
        :type fill_rectangle: bool
        :param border_width: Border width for shapes.
        :type border_width: int
        :param view_height: Fixed view height. None for auto.
        :type view_height: int, optional
        :param widget_id: Unique widget identifier.
        :type widget_id: str, optional
        :param empty_message: Message when no image.
        :type empty_message: str
        """
        self._image_id = None
        super().__init__(
            columns_number=1,
            annotations_opacity=annotations_opacity,
            show_opacity_slider=show_opacity_slider,
            enable_zoom=enable_zoom,
            resize_on_zoom=resize_on_zoom,
            fill_rectangle=fill_rectangle,
            border_width=border_width,
            view_height=view_height,
            empty_message=empty_message,
            widget_id=widget_id,
        )

    def set(
        self,
        title,
        image_url,
        ann: Annotation = None,
        image_id=None,
        zoom_to=None,
        zoom_factor=1.2,
        title_url=None,
        force_clean_up=False,
    ):
        self.clean_up()
        if force_clean_up:
            # if image url is the same, we need to force clean up
            DataJson().send_changes()
        self.append(
            image_url=image_url,
            annotation=ann,
            title=title,
            zoom_to=zoom_to,
            zoom_factor=zoom_factor,
            title_url=title_url,
        )
        self._image_id = image_id
        DataJson().send_changes()

    def clean_up(self):
        super().clean_up()
        self._image_id = None

    def is_empty(self):
        return len(self._data) == 0

    @property
    def id(self):
        return self._image_id



from fastapi import FastAPI, HTTPException, Request, Response
import uvicorn
import threading
import asyncio

from .request_queue import RequestQueue, RequestType
import supervisely as sly
from supervisely import logger


def start_api_server(
    app: sly.Application,
    request_queue: RequestQueue,
    host: str = "0.0.0.0",
    port: int = 8000
) -> threading.Thread:
    """Start FastAPI server in a daemon thread."""
    server = app.get_server()
    create_api(server, request_queue)

    config = uvicorn.Config(app, host=host, port=port, log_level="info")
    server = uvicorn.Server(config)
    
    thread = threading.Thread(target=server.run, daemon=True, name="APIServer")
    thread.start()
    
    logger.debug(f"Live Training API server started: http://{host}:{port}")

    return thread


def create_api(app: FastAPI, request_queue: RequestQueue) -> FastAPI:
    
    @app.post("/start")
    async def start(response: Response):
        """Start the live training process."""
        future = request_queue.put(RequestType.START)
        result = await _wait_for_result(future, response, timeout=None)
        return result
        
    @app.post("/predict")
    async def predict(request: Request, response: Response):
        """Run inference on an image."""
        sly_api = _api_from_request(request)
        state = request.state.state
        img_np = sly_api.image.download_np(state['image_id'])
        future = request_queue.put(
            RequestType.PREDICT,
            {'image': img_np, 'image_id': state['image_id']}
        )
        result = await _wait_for_result(future, response)
        return result
    
    @app.post("/add-sample")
    async def add_sample(request: Request, response: Response):
        """Add a new training sample."""
        sly_api = _api_from_request(request)
        state = request.state.state
        img_np = sly_api.image.download_np(state['image_id'])
        ann_json = sly_api.annotation.download_json(state['image_id'])
        img_info = sly_api.image.get_info_by_id(state['image_id'])
        future = request_queue.put(
            RequestType.ADD_SAMPLE,
            {
                'image': img_np,
                'annotation': ann_json,
                'image_id': state['image_id'],
                'image_name': img_info.name
            }
        )
        result = await _wait_for_result(future, response)
        return result

    @app.post("/status")
    async def status(response: Response):
        """Check the status of the training process."""
        future = request_queue.put(RequestType.STATUS)
        result = await _wait_for_result(future, response)
        return result

    return app


async def _wait_for_result(future: asyncio.Future, response: Response, timeout: float = 600.0):
    """Wait for the future to complete with a timeout."""
    try:
        result = await asyncio.wait_for(future, timeout=timeout)
    except asyncio.TimeoutError:
        # raise HTTPException(503, detail={"error": "Request timeout - training may be busy"})
        response.status_code = 503
        result = _error_response_message("Request timeout - training may be busy")
    except Exception as e:
        # raise HTTPException(500, detail={"error": str(e)})
        response.status_code = 500
        result = _error_response_message(str(e))
    return result


def _api_from_request(request: Request) -> sly.Api:
    api = None
    try:
        api = request.state.api
    finally:
        if not isinstance(api, sly.Api):
            logger.warning("sly.Api instance not found in request.state.api. Creating API from app's credentials.")
            api = sly.Api()
    return api


def _error_response_message(message: str):
    return {"error": {"details": {"message": message}}}


from typing import List

from supervisely.app.content import DataJson, StateJson
from supervisely.app.widgets.widget import Widget


class DropdownCheckboxSelector(Widget):
    """Dropdown widget that lets a user select one or multiple items via checkboxes."""

    class Routes:
        """Callback route names used by the widget frontend to notify Python."""

        VALUE_CHANGED = "value_changed"

    class Item:
        """Selectable item descriptor (id + optional display name/description)."""

        def __init__(self, id: str, name: str = None, description: str = None):
            """
            :param id: Unique item identifier.
            :type id: str
            :param name: Display name. Defaults to id.
            :type name: str, optional
            :param description: Optional description text.
            :type description: str, optional
            """
            self.id = id
            if name is None:
                name = id
            self.name = name
            self.description = description

        def to_json(self):
            return {
                "id": self.id,
                "name": self.name,
                "description": self.description,
            }

        @classmethod
        def from_json(cls, data):
            return cls(
                id=data["id"],
                name=data["name"],
                description=data.get("description", None),
            )

    def __init__(
        self, items: List[Item], label: str = None, widget_id: str = None, multiple: bool = True
    ):
        """

        :param items: List of DropdownCheckboxSelector.Item.
        :type items: List[:class:`~supervisely.app.widgets.dropdown_checkbox_selector.DropdownCheckboxSelector.Item`]
        :param label: Optional label for the dropdown.
        :type label: str, optional
        :param widget_id: Unique widget identifier.
        :type widget_id: str, optional
        :param multiple: If True, allow multiple selection.
        :type multiple: bool, optional
        """
        self._items = items
        self._label = label
        self._multiple = multiple
        self._changes_handled = False

        super().__init__(widget_id=widget_id, file_path=__file__)

    def get_json_data(self):
        return {
            "items": [item.to_json() for item in self._items],
            "label": self._label,
        }

    def get_json_state(self):
        return {
            "selected": [],
        }

    def set(self, items: List[Item]):
        self._items = items
        DataJson()[self.widget_id]["items"] = [
            item.to_json() for item in self._items
        ]
        DataJson().send_changes()

    def get_selected(self) -> List[Item]:
        selected =  StateJson()[self.widget_id].get("selected", [])
        return [item for item in self._items if item.id in selected]

    def select(self, ids: List[str]):
        selected = [item for item in self._items if item.id in ids]
        StateJson()[self.widget_id]["selected"] = [item.to_json() for item in selected]
        StateJson().send_changes()

    def value_changed(self, func):
        """
        Decorator to handle value changes.
        :param func: function to call when value changes
        """

        route_path = self.get_route_path(self.Routes.VALUE_CHANGED)
        server = self._sly_app.get_server()

        @server.post(route_path)
        def on_value_changed():
            selected_items = self.get_selected()
            return func(selected_items)

        self._changes_handled = True

        return on_value_changed



from typing import Dict

from supervisely.nn.benchmark.semantic_segmentation.base_vis_metric import (
    SemanticSegmVisMetric,
)
from supervisely.nn.benchmark.visualization.widgets import ChartWidget, MarkdownWidget


class ClasswiseErrorAnalysis(SemanticSegmVisMetric):
    """Per-class IoU and error metrics chart for semantic segmentation reports."""

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.clickable = True

    @property
    def md(self) -> MarkdownWidget:
        return MarkdownWidget(
            "classwise_error_analysis",
            "Classwise Segmentation Error Analysis",
            text=self.vis_texts.markdown_eou_per_class,
        )

    @property
    def chart(self) -> ChartWidget:
        chart = ChartWidget("classwise_error_analysis", self.get_figure())
        chart.set_click_data(
            self.explore_modal_table.id,
            self.get_click_data(),
            chart_click_extra="'getKey': (payload) => `${payload.points[0].label}`,",
        )
        return chart

    def get_figure(self):
        import plotly.graph_objects as go  # pylint: disable=import-error

        fig = go.Figure()

        # # Classwise Segmentation Error Analysis figure
        bar_data, labels = self.eval_result.mp.classwise_segm_error_data
        color_palette = ["#8ACAA1", "#FFE4B5", "#F7ADAA", "#dd3f3f"]

        for i, column in enumerate(bar_data.columns):
            fig.add_trace(
                go.Bar(
                    name=column,
                    y=bar_data[column],
                    x=labels,
                    marker_color=color_palette[i],
                    hovertemplate="Class: %{x}<br> %{name}: %{y:.2f}<extra></extra>",
                )
            )

        fig.update_layout(barmode="stack", xaxis_title="Class")
        if len(labels) < 10:
            fig.update_layout(width=800)
        fig.update_yaxes(range=[0, 1])

        return fig



from typing import Optional

from supervisely._utils import camel_to_snake, rand_str


class BaseWidget:
    """Base class for benchmark report widgets (chart, table, gallery, etc.); defines save_data, to_html, click data."""

    def __init__(self, name: str = None, title: Optional[str] = None) -> None:
        """:param name: Widget name.
        :param title: Optional display title.
        """
        self.type = camel_to_snake(self.__class__.__name__)
        self.id = f"{self.type}_{rand_str(5)}"
        self.name = name
        self.title = title
        self.data_source = f"/data/{self.name}_{self.id}.json"
        self.click_data_source = f"/data/{self.name}_{self.id}_click_data.json"
        self.switchable = False

    def save_data(self, basepath: str) -> None:
        raise NotImplementedError

    def get_state(self) -> None:
        raise NotImplementedError

    def to_html(self) -> str:
        raise NotImplementedError

    def set_click_data(self, click_gallery_id, click_data) -> None:
        return



from typing import Union
from supervisely.project.project_meta import ProjectMeta
from supervisely.api.api import Api
from supervisely.annotation.annotation import Annotation


# html example
# <sly-grid-gallery
#         v-if="data.gallery"
#         :content="data.gallery.content"
#         :options="data.gallery.options">
# </sly-grid-gallery>


class SingleImageGallery:
    """v1 widget for displaying a single image with optional annotation overlay."""

    def __init__(self, task_id, api: Api, v_model, project_meta: ProjectMeta):
        """
        :param task_id: Task ID.
        :type task_id: int
        :param api: Api instance.
        :type api: :class:`~supervisely.api.api.Api`
        :param v_model: Vue model path.
        :type v_model: str
        :param project_meta: ProjectMeta for annotations.
        :type project_meta: :class:`~supervisely.project.project_meta.ProjectMeta`
        """
        self._task_id = task_id
        self._api = api
        self._v_model = v_model
        self._project_meta = project_meta.clone()
        self._image_url = None
        self._ann = None
        self._options_initialized = False
        self.set_options()

    def set_options(self, enable_zoom=True, show_preview=True, selectable=False, opacity=0.5,
                    show_opacity_in_header=True, fill_rectangle=False):
        """
        :param enable_zoom: If True, enable zoom.
        :type enable_zoom: bool
        :param show_preview: If True, show preview panel.
        :type show_preview: bool
        :param selectable: If True, allow selection.
        :type selectable: bool
        :param opacity: Annotation opacity.
        :type opacity: float
        :param show_opacity_in_header: If True, show opacity control.
        :type show_opacity_in_header: bool
        :param fill_rectangle: If True, fill rectangles.
        :type fill_rectangle: bool
        """
        self._options = {
            "enableZoom": enable_zoom,
            "showPreview": show_preview,
            "selectable": selectable,
            "opacity": opacity,
            "showOpacityInHeader": show_opacity_in_header,
            "fillRectangle": fill_rectangle,
        }
        self._options_initialized = False

    def update_project_meta(self, project_meta: ProjectMeta):
        """
        :param project_meta: ProjectMeta for annotations.
        :type project_meta: :class:`~supervisely.project.project_meta.ProjectMeta`
        """
        self._project_meta = project_meta.clone()

    def set_item(self, image_url, ann: Union[Annotation, dict] = None):
        """
        :param image_url: Image URL.
        :type image_url: str
        :param ann: Annotation.
        :type ann: :class:`~supervisely.annotation.annotation.Annotation`
        """
        self._image_url = image_url
        self.set_annotation(ann)

    def set_annotation(self, ann: Union[Annotation, dict] = None):
        """
        :param ann: Annotation.
        :type ann: :class:`~supervisely.annotation.annotation.Annotation`
        """
        if ann is not None:
            if type(ann) is dict:
                res_ann = Annotation.from_json(ann, self._project_meta)
            else:
                res_ann = ann.clone()
        else:
            res_ann = None
        self._ann = res_ann

    def _get_item_annotation(self):
        figures = [label.to_json() for label in self._ann.labels] if self._ann is not None else None
        return {
            "url": self._image_url,
            "figures": figures,
        }

    def to_json(self):
        if self._image_url is None:
            return None
        return {
            "content": {
                "projectMeta": self._project_meta.to_json(),
                "annotations": {
                    "ann1": self._get_item_annotation(),
                },
                "layout": [["ann1"]]
            },
            "options": {
                **self._options
            }
        }

    def update(self, output=False):
        if self._image_url is None:
            raise ValueError("Image (URL) is not defined")
        content = self.to_json()

        if self._options_initialized is False:
            self._api.task.set_field(self._task_id, self._v_model, content)
            self._options_initialized = True
        else:
            self._api.task.set_field(self._task_id, f"{self._v_model}.content", content["content"])

        if output:
            return content



from __future__ import annotations

from typing import Dict, List, Optional

from supervisely._utils import take_with_default
from supervisely.pointcloud_annotation.pointcloud_episode_object_collection import (
    PointcloudEpisodeObjectCollection,
)
from supervisely.pointcloud_annotation.pointcloud_figure import PointcloudFigure
from supervisely.video_annotation.frame import Frame
from supervisely.video_annotation.key_id_map import KeyIdMap


class PointcloudEpisodeFrame(Frame):
    """Single frame in point cloud episode; holds PointcloudFigures at given index. Immutable."""

    figure_type = PointcloudFigure

    def __init__(self, index: int, figures: Optional[List[PointcloudFigure]] = None):
        """
        Same parameters as :class:`~supervisely.video_annotation.frame.Frame`.

        :param index: Frame index.
        :type index: int
        :param figures: List of PointcloudFigures.
        :type figures: List[:class:`~supervisely.pointcloud_annotation.pointcloud_figure.PointcloudFigure`], optional

        :Usage Example:

            .. code-block:: python

                import supervisely as sly
                from supervisely.geometry.cuboid_3d import Cuboid3d, Vector3d

                obj_class_car = sly.ObjClass('car', Cuboid3d)
                pointcloud_obj_car = sly.PointcloudEpisodeObject(obj_class_car)
                cuboid = Cuboid3d(
                    Vector3d(-3.4, 28.9, -0.7), Vector3d(0., 0, -0.03), Vector3d(1.8, 3.9, 1.6)
                )
                figure = sly.PointcloudFigure(pointcloud_obj_car, cuboid, frame_index=7)
                frame = sly.PointcloudEpisodeFrame(7, figures=[figure])
                print(frame.to_json())
        """
        super().__init__(index, figures)

    @classmethod
    def from_json(
        cls,
        data: Dict,
        objects: PointcloudEpisodeObjectCollection,
        frames_count: Optional[int] = None,
        key_id_map: Optional[KeyIdMap] = None,
    ) -> PointcloudEpisodeFrame:
        """
        Convert a json dict to PointcloudEpisodeFrame. Read more about `Supervisely format <https://docs.supervisely.com/data-organization/00_ann_format_navi>`_.

        :param data: Dict in json format.
        :type data: dict
        :param objects: Pointcloud episode objects collection.
        :type objects: :class:`~supervisely.pointcloud_annotation.pointcloud_episode_object_collection.PointcloudEpisodeObjectCollection`
        :param frames_count: Number of frames in pointcloud episode.
        :type frames_count: int
        :param key_id_map: Key ID map.
        :type key_id_map: :class:`~supervisely.video_annotation.key_id_map.KeyIdMap`
        :raises ValueError: if frame index < 0 and if frame index > number of frames in pointcloud episode
        :returns: Pointcloud episode frame object.
        :rtype: :class:`~supervisely.pointcloud_annotation.pointcloud_episode_frame.PointcloudEpisodeFrame`

        :Usage Example:

            .. code-block:: python

                import supervisely as sly
                from supervisely.geometry.cuboid_3d import Cuboid3d, Vector3d
                from supervisely.pointcloud_annotation.pointcloud_episode_object_collection import PointcloudEpisodeObjectCollection

                obj_class_car = sly.ObjClass('car', Cuboid3d)
                pointcloud_obj_car = sly.PointcloudEpisodeObject(obj_class_car)
                objects = PointcloudEpisodeObjectCollection([pointcloud_obj_car])

                frame_index = 7
                position, rotation, dimension = Vector3d(-3.4, 28.9, -0.7), Vector3d(0., 0, -0.03), Vector3d(1.8, 3.9, 1.6)
                cuboid = Cuboid3d(position, rotation, dimension)

                figure = sly.PointcloudFigure(pointcloud_obj_car, cuboid, frame_index=frame_index)

                frame = sly.PointcloudEpisodeFrame(frame_index, figures=[figure])
                frame_json = frame.to_json()

                frame_from_json = sly.PointcloudEpisodeFrame.from_json(frame_json, objects)
        """

        return super().from_json(data, objects, frames_count, key_id_map)

    def clone(
        self, index: Optional[int] = None, figures: Optional[List[PointcloudFigure]] = None
    ) -> PointcloudEpisodeFrame:
        return super().clone(index, figures)



from typing import Dict, Optional

from supervisely.nn.benchmark.base_visualizer import BaseVisMetric
from supervisely.nn.benchmark.object_detection.evaluator import (
    ObjectDetectionEvalResult,
)


class DetectionVisMetric(BaseVisMetric):
    """Base visualization metric for object/instance detection reports with detection-specific click data."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.eval_result: ObjectDetectionEvalResult

    def get_click_data(self) -> Optional[Dict]:
        if not self.clickable:
            return

        res = {}

        res["layoutTemplate"] = [None, None, None]
        res["clickData"] = {}
        for key, v in self.eval_result.click_data.objects_by_class.items():
            res["clickData"][key] = {}
            res["clickData"][key]["imagesIds"] = []

            img_ids = set()
            obj_ids = set()

            res["clickData"][key][
                "title"
            ] = f"{key} class: {len(v)} object{'s' if len(v) > 1 else ''}"

            for x in v:
                img_ids.add(x["dt_img_id"])
                obj_id = x["dt_obj_id"]
                if obj_id is not None:
                    obj_ids.add(obj_id)

            res["clickData"][key]["imagesIds"] = list(img_ids)
            res["clickData"][key]["filters"] = [
                {
                    "type": "tag",
                    "tagId": "confidence",
                    "value": [self.eval_result.mp.conf_threshold, 1],
                },
                {"type": "tag", "tagId": "outcome", "value": "TP"},
                {"type": "specific_objects", "tagId": None, "value": list(obj_ids)},
            ]

        return res



import os
from typing import List, Tuple

import supervisely.convert.image.label_studio.label_studio_helper as helper
from supervisely import ProjectMeta, logger
from supervisely.annotation.annotation import Annotation
from supervisely.convert.base_converter import AvailableImageConverters
from supervisely.convert.image.image_converter import ImageConverter
from supervisely.io.fs import get_file_ext, get_file_name
from supervisely.io.json import load_json_file


class LabelStudioConverter(ImageConverter):
    """Imports Label Studio export format (JSON with annotations/predictions) into Supervisely image project."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._source_ann_files = []
        self._source_images = {}

    def __str__(self):
        return AvailableImageConverters.LABEL_STUDIO

    @property
    def ann_ext(self) -> str:
        return ".json"

    def validate_key_file(self, ann_path: str) -> bool:
        """Validate LabelMe annotation file."""
        raw_ann = load_json_file(ann_path)
        if not isinstance(raw_ann, list):
            return False
        if len(raw_ann) == 0:
            return False
        if not all([isinstance(ann, dict) for ann in raw_ann]):
            return False
        if not all([isinstance(ann.get("data"), dict) for ann in raw_ann]):
            return False
        anns = []
        for ann in raw_ann:
            anns.extend(ann.get("annotations", []))
            anns.extend(ann.get("predictions", []))
        if not all([isinstance(ann, dict) for ann in anns]):
            return False
        if len(anns) == 0:
            return False
        if not any([ann.get("result") for ann in anns]):
            return False
        return True

    def validate_format(self) -> bool:
        logger.debug(f"Validating format: {self.__str__()}")
        anns_detected = self._find_images_with_annotations()
        if not anns_detected:
            logger.debug(f"No LabelStudio data found in {self._input_data}.")
            return False
        else:
            logger.debug(f"Found {self.items_count} LabelStudio items.")
            return True

    def _find_images_with_annotations(self) -> Tuple[List[ImageConverter.Item], ProjectMeta]:
        """Find images with annotations in Label Studio format."""

        for root, _, files in os.walk(self._input_data):
            for file in files:
                file_name_no_ext = get_file_name(file)
                full_path = os.path.join(root, file)
                ext = get_file_ext(full_path)
                if ext == self.ann_ext:
                    if self.validate_key_file(full_path):
                        self._source_ann_files.append(full_path)
                elif self.is_image(full_path):
                    self._source_images[file_name_no_ext] = full_path

        return len(self._source_ann_files) > 0

    def _create_items(self) -> None:
        """Create items from Label Studio annotations."""

        meta = ProjectMeta()
        items = []
        annotated_images = set()

        for ann_path in self._source_ann_files:
            raw_anns = load_json_file(ann_path)
            for raw_ann in raw_anns:
                annotations = raw_ann.get("annotations", [])
                annotations.extend(raw_ann.get("predictions", []))
                for ann in annotations:
                    image_path = raw_ann.get("data", {}).get("image")
                    if image_path is None:
                        continue
                    name_noext = get_file_name(image_path)
                    image_path = self._source_images.get(name_noext)
                    if image_path is None:
                        continue
                    sly_ann, meta = helper.create_supervisely_annotation(image_path, ann, meta)
                    item = self.Item(image_path, ann_data=sly_ann)
                    items.append(item)
                    if len(sly_ann.labels) > 0 or len(sly_ann.img_tags) > 0:
                        annotated_images.add(name_noext)

        no_ann_images = set(self._source_images.keys()) - annotated_images
        items.extend([self.Item(self._source_images[image_name]) for image_name in no_ann_images])

        self._items = items
        self._meta = meta

    def to_supervisely(
        self,
        item: ImageConverter.Item,
        meta: ProjectMeta = None,
        renamed_classes: dict = None,
        renamed_tags: dict = None,
    ) -> Annotation:
        """Convert to Supervisely format."""

        item.set_shape()
        ann = item.ann_data
        if ann is None:
            return item.create_empty_annotation()
        if renamed_classes:
            new_labels = []
            for label in ann.labels:
                new_cls = renamed_classes.get(label.obj_class.name)
                if new_cls is not None:
                    obj_cls = meta.get_obj_class(new_cls)
                    label = label.clone(obj_class=obj_cls)
                new_labels.append(label)
            ann = ann.clone(labels=new_labels)
        return ann

    def upload_dataset(self, *args, **kwargs):
        """Upload dataset to Supervisely."""

        self._create_items()  # Create items before uploading

        return super().upload_dataset(*args, **kwargs)



from typing import Dict, List

from supervisely.nn.benchmark.visualization.widgets.widget import BaseWidget


class CollapseWidget(BaseWidget):
    """Collapsible container for benchmark report widgets; each child has a foldable panel."""

    def __init__(self, widgets: List[BaseWidget]) -> None:
        """:param widgets: Child widgets with titles for panel headers.
        """
        super().__init__()
        self.widgets = widgets

    def save_data(self, basepath: str) -> None:
        for widget in self.widgets:
            widget.save_data(basepath)

    def get_state(self) -> Dict:
        return {}

    def to_html(self) -> str:
        items = "\n".join(
            [
                f"""
                    <el-collapse-item title="{widget.title}">
                        {widget.to_html()}
                    </el-collapse-item>
                """
                for widget in self.widgets
            ]
        )
        return f"""
                    <el-collapse class='mb-6'>
                        {items}
                    </el-collapse>
                """



from typing import Dict, Optional

from supervisely.app import DataJson
from supervisely.app.widgets import Widget

INFO = "info"
WARNING = "warning"
ERROR = "error"


class DoneLabel(Widget):
    """Displays completion messages (success, done)."""

    def __init__(
        self,
        text: Optional[str] = None,
        widget_id: Optional[str] = None,
    ):
        """
        :param text: Message text.
        :type text: Optional[str]
        :param widget_id: Unique widget identifier.
        :type widget_id: Optional[str]

        :Usage Example:

            .. code-block:: python

                from supervisely.app.widgets import DoneLabel
                label = DoneLabel(text="Done!")
        """
        self._text = text
        super().__init__(widget_id=widget_id, file_path=__file__)

    def get_json_data(self) -> Dict[str, str]:
        """Returns dictionary with widget data, which defines the appearance and behavior of the widget.

        Dictionary contains the following fields:
            - text: DoneLabel text

        :returns: Dictionary with widget data
        :rtype: Dict[str, str]
        """
        return {"text": self._text}

    def get_json_state(self) -> None:
        """DoneLabel widget does not have state,
        the method returns None."""
        return None

    @property
    def text(self) -> str:
        """Returns DoneLabel text.

        :returns: DoneLabel text
        :rtype: str
        """
        return self._text

    @text.setter
    def text(self, value: str) -> None:
        """Sets DoneLabel text.

        :param value: DoneLabel text
        :type value: str
        """
        self._text = value
        DataJson()[self.widget_id]["text"] = self._text
        DataJson().send_changes()



from typing import List

from supervisely.api.api import Api
from supervisely.nn.utils import ModelSource
from supervisely.nn.task_type import TaskType, AVAILABLE_TASK_TYPES
from supervisely.app.widgets.widget import Widget
from supervisely.app.widgets.input.input import Input
from supervisely.app.widgets.field.field import Field
from supervisely.app.widgets.container.container import Container
from supervisely.app.widgets.select.select import Select

class ExternalModelSelector(Widget):
    """Widget for selecting external models by task type, checkpoint path, config, and classes."""

    def __init__(self, task_types: List[str], need_config: bool = True, need_classes: bool = True, widget_id: str = None):
        """
        :param task_types: List of task types, one of: AVAILABLE_TASK_TYPES.
        :type task_types: List[str]
        :param need_config: If True, show config path input.
        :type need_config: bool
        :param need_classes: If True, show classes file input.
        :type need_classes: bool
        :param widget_id: Unique widget identifier.
        :type widget_id: str, optional
        """
        self.need_config = need_config
        self.need_classes = need_classes
        self.widgets = []

        # Model name input
        self.model_name_input = Input(value="External Model", placeholder="Enter model name")
        self.model_name_field = Field(
            title="Model name",
            description="Model name is used to identify the model in model benchmark and comparison.",
            content=self.model_name_input,
        )
        self.widgets.append(self.model_name_field)

        # Task type selector
        if len(task_types) == 0:
            raise ValueError("At least 1 task type must be provided for external models")
        for task_type in task_types:
            if task_type not in AVAILABLE_TASK_TYPES:
                raise ValueError(f"Invalid task type: {task_type}. Available task types: {AVAILABLE_TASK_TYPES}")

        self.task_types = task_types

        if len(task_types) > 1:
            select_items = [
                Select.Item(value=task_type, label=task_type) for task_type in task_types
            ]
            self.task_type_selector = Select(items=select_items)
            self.task_type_field = Field(
                title="Task type",
                description="Task type of the model",
                content=self.task_type_selector,
            )
            self.widgets.append(self.task_type_field)

        # Checkpoint input
        self.checkpoint_input = Input(placeholder="Enter checkpoint path")
        self.checkpoint_field = Field(title="Checkpoint", description="Path to the checkpoint in Team Files", content=self.checkpoint_input)
        self.widgets.append(self.checkpoint_field)

        # Config input
        if need_config:
            self.config_input = Input(placeholder="Enter config path")
            self.config_field = Field(
                title="Config",
                description="Path to the model config in Team Files",
                content=self.config_input,
            )
            self.widgets.append(self.config_field)

        # Classes input
        if need_classes:
            self.classes_input = Input(placeholder="Enter classes")
            self.classes_field = Field(
                title="Model classes",
                description="Path to '.json' file with classes in Team Files. File should contain classes mapping (e.g {'0': 'person', '1': 'car', '2': 'pizza'})",
                content=self.classes_input,
            )
            self.widgets.append(self.classes_field)

        # Layout
        self.layout = Container(self.widgets)

        super().__init__(widget_id=widget_id)

    def get_json_data(self):
        return {}

    def get_json_state(self):
        return {
            "modelName": self.get_model_name(),
            "taskType": self.get_task_type(),
            "checkpointPath": self.get_checkpoint_path(),
            "configPath": self.get_config_path(),
            "classesPath": self.get_classes_path(),
        }

    def get_model_name(self):
        return self.model_name_input.get_value()

    def set_model_name(self, model_name: str):
        self.model_name_input.set_value(model_name)

    def get_task_type(self):
        if len(self.task_types) == 1:
            return self.task_types[0]
        return self.task_type_selector.get_value()

    def set_task_type(self, task_type: str):
        self.task_type_selector.set_value(task_type)

    def get_checkpoint_path(self):
        return self.checkpoint_input.get_value()

    def set_checkpoint_path(self, checkpoint_path: str):
        self.checkpoint_input.set_value(checkpoint_path)

    def get_config_path(self):
        if self.need_config:
            return self.config_input.get_value()
        return None

    def set_config_path(self, config_path: str):
        if self.need_config:
            self.config_input.set_value(config_path)
        else:
            raise ValueError("Enable config path is required")

    def get_classes_path(self):
        if self.need_classes:
            return self.classes_input.get_value()
        return None

    def set_classes_path(self, classes_path: str):
        if self.need_classes:
            self.classes_input.set_value(classes_path)
        else:
            raise ValueError("Enable classes path is required")

    def get_deploy_params(self):
        model_info = {"model_name": self.get_model_name(), "task_type": self.get_task_type()}
        model_files = {"checkpoint": self.get_checkpoint_path()}
        if self.need_config:
            model_files["config"] = self.get_config_path()
        if self.need_classes:
            model_files["classes"] = self.get_classes_path()

        deploy_params = {
            "model_info": model_info,
            "model_files": model_files,
            "model_source": ModelSource.EXTERNAL,
        }
        return deploy_params

    def to_html(self):
        return self.layout.to_html()



from typing import List, Tuple

from supervisely.annotation.annotation import Annotation
from supervisely.api.image_api import ImageInfo
from supervisely.api.module_api import ApiField
from supervisely.nn.benchmark.base_visualizer import BaseVisMetrics
from supervisely.nn.benchmark.visualization.widgets import GalleryWidget, MarkdownWidget
from supervisely.project.project_meta import ProjectMeta


class ExplorePredictions(BaseVisMetrics):
    """Gallery and widgets to explore prediction differences across compared segmentation models."""


    MARKDOWN = "markdown_explorer"
    GALLERY_DIFFERENCE = "explore_difference_gallery"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.meta = None

    def _merged_meta(self) -> ProjectMeta:
        if self.meta is not None:
            return self.meta
        self.meta = self.eval_results[0].gt_project_meta
        for eval_res in self.eval_results:
            self.meta = self.meta.merge(eval_res.pred_project_meta)
        return self.meta

    @property
    def difference_predictions_md(self) -> MarkdownWidget:
        text = self.vis_texts.markdown_explorer
        return MarkdownWidget(self.MARKDOWN, "Explore Predictions", text)

    @property
    def explore_gallery(self) -> GalleryWidget:
        columns_number = len(self.eval_results) + 1
        images, annotations = self._get_sample_data()
        gallery = GalleryWidget(self.GALLERY_DIFFERENCE, columns_number=columns_number)
        gallery.add_image_left_header("Click to explore more")
        gallery.set_project_meta(self._merged_meta())
        gallery.set_images(images, annotations)
        click_data = self.get_click_data_explore_all()
        gallery.set_click_data(self.explore_modal_table.id, click_data)
        gallery.set_show_all_data(self.explore_modal_table.id, click_data)
        gallery._gallery._update_filters()

        return gallery

    def _get_sample_data(self) -> Tuple[List[ImageInfo], List[Annotation], List[ProjectMeta]]:
        images = []
        annotations = []
        api = self.eval_results[0].api
        names = None
        ds_name = None
        for idx, eval_res in enumerate(self.eval_results):
            if idx == 0:
                dataset_info = eval_res.gt_dataset_infos[0]
                infos = api.image.get_list(dataset_info.id, limit=5, force_metadata_for_links=False)
                ds_name = dataset_info.name
                images_ids = [image_info.id for image_info in infos]
                names = [image_info.name for image_info in infos]
                images.append(infos)
                from supervisely.api.api import Api

                api: Api
                anns = api.annotation.download_batch(
                    dataset_info.id, images_ids, force_metadata_for_links=False
                )
                annotations.append(anns)
            assert ds_name is not None, "Failed to get GT dataset name for gallery"

            dataset_info = api.dataset.get_info_by_name(eval_res.pred_project_id, ds_name)

            assert names is not None, "Failed to get GT image names for gallery"
            infos = api.image.get_list(
                dataset_info.id,
                filters=[
                    {ApiField.FIELD: ApiField.NAME, ApiField.OPERATOR: "in", ApiField.VALUE: names}
                ],
                force_metadata_for_links=False,
            )
            images_ids = [image_info.id for image_info in infos]
            images.append(infos)
            anns = api.annotation.download_batch(
                dataset_info.id, images_ids, force_metadata_for_links=False
            )
            annotations.append(anns)

        images = list(i for x in zip(*images) for i in x)
        annotations = list(i for x in zip(*annotations) for i in x)
        return images, annotations

    def get_click_data_explore_all(self) -> dict:
        res = {}

        res["projectMeta"] = self._merged_meta().to_json()
        res["layoutTemplate"] = [{"columnTitle": "Ground Truth"}]
        for idx, eval_res in enumerate(self.eval_results, 1):
            res["layoutTemplate"].append({"columnTitle": f"[{idx}] {eval_res.short_name}"})

        click_data = res.setdefault("clickData", {})
        explore = click_data.setdefault("explore", {})
        explore["title"] = "Explore all predictions"

        image_names = set()
        for eval_res in self.eval_results:
            eval_res.mp.per_image_metrics["img_names"].apply(image_names.add)

        filters = [{"field": "name", "operator": "in", "value": list(image_names)}]

        images_ids = []
        api = self.eval_results[0].api
        names = None
        ds_names = None
        for idx, eval_res in enumerate(self.eval_results):
            if idx == 0:
                dataset_infos = eval_res.gt_dataset_infos
                ds_names = [ds.name for ds in dataset_infos]
                current_images_ids = []
                current_images_names = []
                for ds in dataset_infos:
                    image_infos = api.image.get_list(ds.id, filters, force_metadata_for_links=False)
                    image_infos = sorted(image_infos, key=lambda x: x.name)
                    current_images_names.extend([image_info.name for image_info in image_infos])
                    current_images_ids.extend([image_info.id for image_info in image_infos])
                images_ids.append(current_images_ids)
                names = current_images_names

            dataset_infos = api.dataset.get_list(eval_res.pred_project_id)
            dataset_infos = [ds for ds in dataset_infos if ds.name in ds_names]
            dataset_infos = sorted(dataset_infos, key=lambda x: ds_names.index(x.name))
            current_images_infos = []
            for ds in dataset_infos:
                image_infos = api.image.get_list(ds.id, filters, force_metadata_for_links=False)
                image_infos = [image_info for image_info in image_infos if image_info.name in names]
                current_images_infos.extend(image_infos)
            current_images_infos = sorted(current_images_infos, key=lambda x: names.index(x.name))
            images_ids.append([image_info.id for image_info in current_images_infos])

        explore["imagesIds"] = list(i for x in zip(*images_ids) for i in x)

        return res



# coding: utf-8
"""API for working with Webhooks"""

from __future__ import annotations

from typing import TYPE_CHECKING, Dict, List, NamedTuple, Optional

from supervisely.api.module_api import ApiField, ModuleNoParent

if TYPE_CHECKING:
    from supervisely.api.api import Api


# Available webhook action types
LABELING_JOB_COMPLETED = "labeling_job.completed"
LABELING_QUEUE_JOB_COMPLETED = "labeling_queue.job.completed"
LABELING_QUEUE_COMPLETED = "labeling_queue.completed"


class WebhookInfo(NamedTuple):
    """Information about a webhook returned from API"""

    id: int
    """Webhook ID"""
    action: str
    """Webhook action type"""
    url: str
    """Target URL"""
    retries_count: int
    """Number of retry attempts"""
    retries_delay: int
    """Delay between retries in seconds"""
    team_id: int
    """Team ID"""
    created_at: str
    """Creation timestamp"""
    updated_at: str
    """Last update timestamp"""


class WebhookApi(ModuleNoParent):
    """API for working with webhooks."""

    def __init__(self, api):
        """
        :param api: :class:`~supervisely.api.api.Api` object to use for API connection.
        :type api: :class:`~supervisely.api.api.Api`

        :Usage Example:

            .. code-block:: python

                import supervisely as sly
                api = sly.Api.from_env()
                webhook = api.webhook.create(
                    team_id=123,
                    url="https://example.com/webhook",
                    action=sly.LABELING_JOB_COMPLETED,
                )
        """
        super().__init__(api)

    def get_list(
        self, team_id: int, filters: Optional[List[Dict]] = None
    ) -> List[WebhookInfo]:
        """
        Get list of all webhooks in a team.

        :param team_id: Team ID
        :type team_id: int
        :param filters: List of params to filter webhooks
        :type filters: List[Dict], optional
        :return: List of webhook information objects
        :rtype: List[WebhookInfo]
        :Usage example:

         .. code-block:: python

            import supervisely as sly

            api = sly.Api.from_env()
            webhooks = api.webhook.get_list(team_id=123)
        """
        data = {ApiField.TEAM_ID: team_id}
        if filters is not None:
            data[ApiField.FILTER] = filters
        response = self._api.post("webhooks.list", data)
        results = []
        for item in response.json().get("entities", []):
            meta = item.get("meta", {})
            results.append(WebhookInfo(
                id=item.get(ApiField.ID),
                action=item.get("action"),
                url=item.get(ApiField.URL),
                retries_count=meta.get("retriesCount", 5),
                retries_delay=meta.get("retriesDelay", 10),
                team_id=item.get(ApiField.TEAM_ID),
                created_at=item.get(ApiField.CREATED_AT),
                updated_at=item.get(ApiField.UPDATED_AT),
            ))
        return results

    def get_info_by_id(self, webhook_id: int) -> WebhookInfo:
        """
        Get webhook information by ID.

        :param webhook_id: Webhook ID
        :type webhook_id: int
        :return: Webhook information
        :rtype: WebhookInfo
        :Usage example:

         .. code-block:: python

            import supervisely as sly

            api = sly.Api.from_env()
            webhook_info = api.webhook.get_info_by_id(webhook_id=456)
        """
        response = self._api.post("webhooks.info", {ApiField.ID: webhook_id})
        info = response.json()
        meta = info.get("meta", {})
        return WebhookInfo(
            id=info.get(ApiField.ID),
            action=info.get("action"),
            url=info.get(ApiField.URL),
            retries_count=meta.get("retriesCount", 5),
            retries_delay=meta.get("retriesDelay", 10),
            team_id=info.get(ApiField.TEAM_ID),
            created_at=info.get(ApiField.CREATED_AT),
            updated_at=info.get(ApiField.UPDATED_AT),
        )

    def create(
        self,
        team_id: int,
        url: str,
        action: str,
        retries_count: int = 5,
        retries_delay: int = 10,
        headers: Optional[Dict] = None,
        follow_redirect: bool = True,
        reject_unauthorized: bool = True,
    ) -> WebhookInfo:
        """
        Create a new webhook.

        :param team_id: Team ID
        :type team_id: int
        :param url: Target URL
        :type url: str
        :param action: Webhook action type (use WebhookAction constants)
        :type action: str
        :param retries_count: Number of retry attempts
        :type retries_count: int
        :param retries_delay: Delay between retries in seconds
        :type retries_delay: int
        :param headers: Custom headers
        :type headers: Dict, optional
        :param follow_redirect: Whether to follow redirects
        :type follow_redirect: bool
        :param reject_unauthorized: Whether to reject unauthorized SSL
        :type reject_unauthorized: bool
        :return: Created webhook information
        :rtype: WebhookInfo
        :Usage example:

         .. code-block:: python

            import supervisely as sly

            api = sly.Api.from_env()
            webhook = api.webhook.create(
                team_id=123,
                url="https://example.com/webhook",
                action=sly.WebhookAction.LABELING_JOB_COMPLETED,
                retries_count=3
            )
        """
        if headers is None:
            headers = {}

        meta = {
            "retriesCount": retries_count,
            "retriesDelay": retries_delay,
            "headers": headers,
            "followRedirect": follow_redirect,
            "rejectUnauthorized": reject_unauthorized,
        }

        data = {
            ApiField.TEAM_ID: team_id,
            ApiField.URL: url,
            "action": action,
            "meta": meta,
        }

        response = self._api.post("webhooks.add", data)
        webhook_id = response.json()["id"]
        return self.get_info_by_id(webhook_id)

    def update(
        self,
        webhook_id: int,
        url: Optional[str] = None,
        action: Optional[str] = None,
        retries_count: Optional[int] = None,
        retries_delay: Optional[int] = None,
        headers: Optional[Dict] = None,
        follow_redirect: Optional[bool] = None,
        reject_unauthorized: Optional[bool] = None,
    ) -> WebhookInfo:
        """
        Update an existing webhook.

        :param webhook_id: Webhook ID to update
        :type webhook_id: int
        :param url: New target URL
        :type url: str, optional
        :param action: New action type
        :type action: str, optional
        :param retries_count: New number of retry attempts
        :type retries_count: int, optional
        :param retries_delay: New delay between retries
        :type retries_delay: int, optional
        :param headers: New custom headers
        :type headers: Dict, optional
        :param follow_redirect: New follow redirect setting
        :type follow_redirect: bool, optional
        :param reject_unauthorized: New reject unauthorized setting
        :type reject_unauthorized: bool, optional
        :return: Updated webhook information
        :rtype: WebhookInfo
        :Usage example:

         .. code-block:: python

            import supervisely as sly

            api = sly.Api.from_env()
            updated = api.webhook.update(webhook_id=456, url="https://new.com/webhook")
        """
        data = {ApiField.ID: webhook_id}

        if url is not None:
            data[ApiField.URL] = url
        if action is not None:
            data["action"] = action

        # Build meta object if any meta fields are provided
        meta_fields = {}
        if retries_count is not None:
            meta_fields["retriesCount"] = retries_count
        if retries_delay is not None:
            meta_fields["retriesDelay"] = retries_delay
        if headers is not None:
            meta_fields["headers"] = headers
        if follow_redirect is not None:
            meta_fields["followRedirect"] = follow_redirect
        if reject_unauthorized is not None:
            meta_fields["rejectUnauthorized"] = reject_unauthorized

        if meta_fields:
            data["meta"] = meta_fields

        response = self._api.post("webhooks.update", data)
        webhook_id = response.json()["id"]
        return self.get_info_by_id(webhook_id)

    def remove(self, webhook_id: int) -> None:
        """
        Remove a webhook by ID.

        :param webhook_id: Webhook ID to remove
        :type webhook_id: int
        :Usage example:

         .. code-block:: python

            import supervisely as sly

            api = sly.Api.from_env()
            api.webhook.remove(webhook_id=456)
        """
        data = {"ids": [webhook_id]}
        self._api.post("webhooks.bulk.remove", data)

    def test(
        self,
        team_id: int,
        webhook_id: Optional[int] = None,
        action: Optional[str] = None,
        payload: Optional[Dict] = None,
    ) -> Dict:
        """
        Send a test event to a webhook.

        :param team_id: Team ID
        :type team_id: int
        :param webhook_id: Webhook ID to test
        :type webhook_id: int, optional
        :param action: Action type to simulate
        :type action: str, optional
        :param payload: Custom payload to send
        :type payload: Dict, optional
        :return: Response from the test
        :rtype: Dict
        :Usage example:

         .. code-block:: python

            import supervisely as sly

            api = sly.Api.from_env()
            response = api.webhook.test(team_id=123, webhook_id=456)
        """
        data = {ApiField.TEAM_ID: team_id}

        if webhook_id is not None:
            data[ApiField.ID] = webhook_id
        if action is not None:
            data["action"] = action
        if payload is not None:
            data["payload"] = payload
        else:
            data["payload"] = {}

        response = self._api.post("webhooks.test", data)
        return response.json()



import copy
import functools
from enum import Enum

try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal

import fastapi
from varname import varname

from supervisely.app import DataJson
from supervisely.app.fastapi import run_sync
from supervisely.app.widgets import Widget


class ElementButton(Widget):
    """Button widget based on Element UI styling that can be updated (text/loading/disabled) from Python."""

    class Routes:
        """Callback route names used by the widget frontend to notify Python."""

        BUTTON_CLICKED = "button_clicked_cb"

    def __init__(
        self,
        text: str = "Element Button",
        button_type: Literal[
            "primary", "info", "warning", "danger", "success"
        ] = "primary",
        button_size: Literal["mini", "small", "large"] = None,
        plain: bool = False,
        widget_id: str = None,
    ):
        """
        :param text: Text to display on the button.
        :type text: str
        :param button_type: Button type: "primary", "info", "warning", "danger", "success".
        :type button_type: Literal["primary", "info", "warning", "danger", "success"]
        :param button_size: Button size: "mini", "small", or "large".
        :type button_size: Literal["mini", "small", "large"], optional
        :param plain: If True, use plain (outline) style.
        :type plain: bool
        :param widget_id: Unique widget identifier.
        :type widget_id: str, optional
        """
        self._widget_routes = {}

        self._text = text
        self._button_type = button_type
        self._plain = plain
        self._button_size = button_size

        self._loading = False
        self._disabled = False

        super().__init__(widget_id=widget_id, file_path=__file__)

    def get_json_data(self):
        return {
            "text": self._text,
            "button_type": self._button_type,
            "plain": self._plain,
            "button_size": self._button_size,
            "loading": self._loading,
            "disabled": self._disabled,
        }

    def get_json_state(self):
        return None

    @property
    def text(self):
        return self._text

    @text.setter
    def text(self, value):
        self._text = value
        DataJson()[self.widget_id]["text"] = self._text

    @property
    def loading(self):
        return self._loading

    @loading.setter
    def loading(self, value):
        self._loading = value
        DataJson()[self.widget_id]["loading"] = self._loading

    @property
    def disabled(self):
        return self._disabled

    @disabled.setter
    def disabled(self, value):
        self._disabled = value
        DataJson()[self.widget_id]["disabled"] = self._disabled

    def click(self):
        route = ElementButton.Routes.BUTTON_CLICKED
        return self.add_route(self._sly_app.get_server(), route)



# coding: utf-8

from supervisely.geometry.rectangle import Rectangle
from supervisely.geometry.validation import is_2d_int_coords_valid
from supervisely.geometry.sliding_windows import SlidingWindows
from supervisely.collection.str_enum import StrEnum


class SlidingWindowBorderStrategy(StrEnum):
    """Strategy for handling border of the sliding window."""

    ADD_PADDING = "add_padding"
    """Strategy for adding padding to the sliding window."""
    SHIFT_WINDOW = "shift_window"
    """Strategy for shifting the sliding window."""
    CHANGE_SIZE = "change_size"
    """Strategy for changing the size of the sliding window."""


class SlidingWindowsFuzzy(SlidingWindows):
    """Sliding windows with fuzzy border strategy."""

    def __init__(
        self, window_shape, min_overlap, strategy=str(SlidingWindowBorderStrategy.SHIFT_WINDOW)
    ):
        """
        Sliding windows with fuzzy border strategy.

        :param window_shape: Shape of the window.
        :type window_shape: tuple
        :param min_overlap: Minimum overlap of the window.
        :type min_overlap: tuple
        :param strategy: Strategy for handling border of the sliding window.
        :type strategy: :class:`~supervisely.geometry.sliding_windows_fuzzy.SlidingWindowBorderStrategy`
        """
        super().__init__(window_shape, min_overlap)
        if not SlidingWindowBorderStrategy.has_value(strategy):
            raise ValueError(
                "Unknown strategy {!r}. Allowed strategies: {}".format(
                    strategy, SlidingWindowBorderStrategy.values()
                )
            )
        self.strategy = strategy

    def get(self, source_shape):
        """Get the sliding windows."""
        if self.strategy == str(SlidingWindowBorderStrategy.SHIFT_WINDOW):
            yield from self.get_shift_window(source_shape)
        elif self.strategy == str(SlidingWindowBorderStrategy.ADD_PADDING):
            yield from self.get_add_padding(source_shape)
        elif self.strategy == str(SlidingWindowBorderStrategy.CHANGE_SIZE):
            yield from self.get_change_size(source_shape)
        else:
            raise NotImplementedError("Not implemented SW Strategy: {!r}".format(self.strategy))

    def get_shift_window(self, source_shape):
        """Get the sliding windows with shift window strategy."""
        yield from super().get(source_shape)

    def get_add_padding(self, source_shape):
        """Get the sliding windows with add padding strategy."""
        h = source_shape[0]
        w = source_shape[1]
        source_rect = Rectangle.from_size(source_shape)
        window_rect = Rectangle.from_size(self.window_shape)
        if not source_rect.contains(window_rect):
            raise RuntimeError("Sliding window: window is larger than source (image).")

        # hw_limit = tuple(source_shape[i] - self.window_shape[i] for i in (0, 1))
        # for wind_top in range(0, hw_limit[0] + self.stride[0], self.stride[0]):
        # for wind_left in range(0, hw_limit[1] + self.stride[1], self.stride[1]):
        for wind_top in range(0, h, self.stride[0]):
            for wind_left in range(0, w, self.stride[1]):
                roi = window_rect.translate(drow=wind_top, dcol=wind_left)
                yield roi

    def get_change_size(self, source_shape):
        """Get the sliding windows with change size strategy."""
        h = source_shape[0]
        w = source_shape[1]
        source_rect = Rectangle.from_size(source_shape)
        window_rect = Rectangle.from_size(self.window_shape)
        if not source_rect.contains(window_rect):
            raise RuntimeError("Sliding window: window is larger than source (image).")

        # hw_limit = tuple(source_shape[i] - self.window_shape[i] for i in (0, 1))
        # for wind_top in range(0, hw_limit[0] + self.stride[0], self.stride[0]):
        #     for wind_left in range(0, hw_limit[1] + self.stride[1], self.stride[1]):
        for wind_top in range(0, h, self.stride[0]):
            for wind_left in range(0, w, self.stride[1]):
                wind_bottom = min(wind_top + self.stride[0], source_shape[0])
                wind_right = min(wind_left + self.stride[1], source_shape[1])
                roi = Rectangle(wind_top, wind_left, wind_bottom - 1, wind_right - 1)
                if not source_rect.contains(roi):
                    raise RuntimeError("Sliding window: result crop bounds are invalid.")
                yield roi



# coding: utf-8

# docs
from typing import Callable, Dict, List, Optional, Tuple, Union

from tqdm import tqdm

from supervisely._utils import batched
from supervisely.api.module_api import ApiField, ModuleApiBase


class AdvancedApi(ModuleApiBase):
    """Advanced API for working with images and objects."""

    def add_tag_to_object(
        self, tag_meta_id: int, figure_id: int, value: Optional[Union[str, int]] = None
    ) -> Dict:
        """Add a tag to an object.

        :param tag_meta_id: Tag meta ID.
        :type tag_meta_id: int
        :param figure_id: Figure ID.
        :type figure_id: int
        :param value: Value of the tag. If not provided, the tag will be added without a value.
        :type value: Optional[Union[str, int]]
        :returns: Dictionary with the result.
        :rtype: dict
        """
        data = {ApiField.TAG_ID: tag_meta_id, ApiField.FIGURE_ID: figure_id}
        if value is not None:
            data[ApiField.VALUE] = value
        resp = self._api.post("object-tags.add-to-object", data)
        return resp.json()

    def remove_tag_from_object(self, tag_meta_id: int, figure_id: int, tag_id: int) -> Dict:
        """
        Remove a tag from an object.

        :param tag_meta_id: Tag meta ID.
        :type tag_meta_id: int
        :param figure_id: Figure ID.
        :type figure_id: int
        :param tag_id: Tag ID.
        :type tag_id: int
        :returns: Dictionary with the result.
        :rtype: dict
        """
        data = {ApiField.TAG_ID: tag_meta_id, ApiField.FIGURE_ID: figure_id, ApiField.ID: tag_id}
        resp = self._api.post("object-tags.remove-from-figure", data)
        return resp.json()

    def get_object_tags(self, figure_id: int) -> Dict:
        """Get tags for an object.

        :param figure_id: Figure ID.
        :type figure_id: int
        :returns: Dictionary with the result.
        :rtype: dict
        """
        data = {ApiField.ID: figure_id}
        resp = self._api.post("figures.tags.list", data)
        return resp.json()

    def remove_tag_from_image(self, tag_meta_id: int, image_id: int, tag_id: int) -> Dict:
        """Remove a tag from an image.

        :param tag_meta_id: Tag meta ID.
        :type tag_meta_id: int
        :param image_id: Image ID.
        :type image_id: int
        :param tag_id: Tag ID.
        :type tag_id: int
        :returns: Dictionary with the result.
        :rtype: dict
        """
        data = {ApiField.TAG_ID: tag_meta_id, ApiField.IMAGE_ID: image_id, ApiField.ID: tag_id}
        resp = self._api.post("image-tags.remove-from-image", data)
        return resp.json()

    def remove_tags_from_images(
        self,
        tag_meta_ids: List[int],
        image_ids: List[int],
        progress_cb: Optional[Union[tqdm, Callable]] = None,
    ) -> None:
        """
        Remove tags from images.

        :param tag_meta_ids: List of tag meta IDs.
        :type tag_meta_ids: List[int]
        :param image_ids: List of image IDs.
        :type image_ids: List[int]
        :param progress_cb: Function for control remove progress.
        :type progress_cb: Callable
        :returns: None
        :rtype: None
        """
        for batch_ids in batched(image_ids, batch_size=100):
            data = {ApiField.TAG_IDS: tag_meta_ids, ApiField.IDS: batch_ids}
            self._api.post("image-tags.bulk.remove-from-images", data)
            if progress_cb is not None:
                progress_cb(len(batch_ids))



import json
from pathlib import Path
from typing import Any, List, Optional, Union

from jinja2 import Template

from supervisely.api.annotation_api import AnnotationInfo
from supervisely.api.image_api import ImageInfo
from supervisely.io.fs import ensure_base_path
from supervisely.nn.benchmark.visualization.widgets.widget import BaseWidget
from supervisely.project.project_meta import ProjectMeta


class GalleryWidget(BaseWidget):
    """Benchmark report widget that displays image gallery with annotations and click-to-inspect binding."""

    def __init__(
        self,
        name: str,
        filters: Optional[List] = None,
        is_modal: Optional[bool] = False,
        columns_number: int = 3,
        click_gallery_id: Optional[str] = None,
        opacity: Optional[float] = 0.4,
        limit: Optional[int] = None,
    ):
        """:param name: Widget name.
        :param filters: Optional tag filters.
        :param is_modal: Show as modal.
        :param columns_number: Gallery columns.
        :param click_gallery_id: Gallery ID for click handler.
        :param opacity: Annotation opacity.
        :param limit: Max items in click gallery.
        """
        super().__init__(name)
        self.reference = self.id
        self.is_modal = is_modal
        self.click_handled = False
        self.click_gallery_id = click_gallery_id
        self.click_data = None
        self.click_gallery_items_limit = limit or columns_number * 3
        self.image_left_header = False
        self._project_meta = None
        self.show_all_button = False
        self.show_all_gallery_id = None
        self.show_all_data = None
        self.show_all_data_source = f"/data/{self.name}_{self.id}_show_all_data.json"
        self.columns_number = columns_number
        self.get_key = None
        self.opacity = opacity or 0.4

        filters = filters  # or [{"confidence": [0.6, 1]}]

        from supervisely.app.widgets import GridGalleryV2

        self._gallery = GridGalleryV2(
            columns_number=columns_number,
            annotations_opacity=self.opacity,
            border_width=4,
            enable_zoom=False,
            default_tag_filters=filters,
            show_zoom_slider=False,
        )

    def set_project_meta(self, project_meta: ProjectMeta):
        self._project_meta = project_meta

    def set_images(
        self,
        image_infos: List[ImageInfo],
        ann_infos: List[AnnotationInfo],
        project_metas: List[ProjectMeta] = None,
        skip_tags_filtering: Optional[List[Union[bool, List[str]]]] = None,
    ):
        """One time operation"""
        if project_metas is None:
            project_metas = [self._project_meta] * self.columns_number

        if skip_tags_filtering is None:
            skip_tags_filtering = [False] * self.columns_number

        for idx, (image, ann) in enumerate(zip(image_infos, ann_infos)):
            image_name = image.name
            image_url = image.full_storage_url
            self._gallery.append(
                title=image_name,
                image_url=image_url,
                annotation_info=ann,
                column_index=idx % self.columns_number,
                project_meta=project_metas[idx % self.columns_number],
                ignore_tags_filtering=skip_tags_filtering[idx % self.columns_number],
                call_update=idx == len(image_infos) - 1,
            )

    def _get_init_data(self):
        res = {}
        self._gallery._update_filters()
        res.update(self._gallery.get_json_state())
        res.update(self._gallery.get_json_data()["content"])
        res["layoutData"] = res.pop("annotations")
        res["projectMeta"] = self._project_meta.to_json()
        return res

    def save_data(self, basepath: str) -> None:
        # init data
        basepath = basepath.rstrip("/")
        ensure_base_path(basepath + self.data_source)

        with open(basepath + self.data_source, "w") as f:
            json.dump(self._get_init_data(), f)

        # click data
        if self.click_data is not None:
            ensure_base_path(basepath + self.click_data_source)
            with open(basepath + self.click_data_source, "w") as f:
                json.dump(self.click_data, f)

        # show all data
        if self.show_all_data is not None:
            ensure_base_path(basepath + self.show_all_data_source)
            with open(basepath + self.show_all_data_source, "w") as f:
                json.dump(self.show_all_data, f)

    def get_state(self) -> None:
        return {}

    def to_html(self) -> str:
        template_str = Path(__file__).parent / "template.html"
        return Template(template_str.read_text()).render(self._get_template_data())

    def set_click_data(
        self, click_gallery_id: str, click_data: Any, get_key: Optional[str] = None
    ) -> None:
        self.click_handled = True
        self.click_data = click_data
        self.click_gallery_id = click_gallery_id
        if get_key is not None:
            self.get_key = get_key

    def set_show_all_data(self, gallery_id: str, data: str) -> None:
        self.show_all_button = True
        self.show_all_gallery_id = gallery_id
        self.show_all_data = data

    def add_image_left_header(self, html: str):
        self.image_left_header = html

    def _get_template_data(self):
        return {
            "widget_id": self.id,
            "reference": self.reference,  # TODO: same as id?
            "init_data_source": self.data_source,
            "is_modal": str(self.is_modal).lower(),
            "click_handled": self.click_handled,
            "show_all_button": self.show_all_button,
            "show_all_gallery_id": self.show_all_gallery_id,
            "show_all_data_source": self.show_all_data_source,
            "click_data_source": self.click_data_source,
            "click_gallery_id": self.click_gallery_id,
            "click_gallery_items_limit": self.click_gallery_items_limit,
            "image_left_header": self.image_left_header,
            "get_key": self.get_key,
        }



from supervisely.collection.str_enum import StrEnum


class CVTask(StrEnum):
    """Enum of supported computer vision task types for benchmarking."""

    OBJECT_DETECTION: str = "object_detection"
    INSTANCE_SEGMENTATION: str = "instance_segmentation"
    SEMANTIC_SEGMENTATION: str = "semantic_segmentation"



from __future__ import annotations

from typing import Any, Callable, Dict, Optional, Union

from supervisely.app import StateJson
from supervisely.app.widgets import Text, Widget


class Checkbox(Widget):
    """Checkbox for selecting an option."""

    class Routes:
        """Route name constants for this widget."""
        VALUE_CHANGED = "value_changed"

    def __init__(
        self,
        content: Union[Widget, str],
        checked: Optional[bool] = False,
        widget_id: Optional[str] = None,
    ):
        """
        :param content: Label widget or text string.
        :type content: Union[Widget, str]
        :param checked: Initial checked state.
        :type checked: bool, optional
        :param widget_id: Widget identifier.
        :type widget_id: str, optional

        :Usage Example:

            .. code-block:: python

                from supervisely.app.widgets import Checkbox
                cb = Checkbox("Enable feature", checked=False)
        """
        self._content = content
        self._checked = checked
        if type(self._content) is str:
            self._content = [Text(self._content)][0]
        self._changes_handled = False
        super().__init__(widget_id=widget_id, file_path=__file__)

    def get_json_data(self) -> Dict:
        """Checkbox widget does not have any additional data, so it returns an empty dictionary.

        :returns: Empty dictionary.
        :rtype: Dict
        """
        return {}

    def get_json_state(self) -> Dict[str, bool]:
        """Returns the state of the checkbox.
        Dictionary contains the following keys:
            - checked: Current state of the checkbox.

        :returns: Dictionary with the state of the checkbox.
        :rtype: Dict[str, bool]
        """
        return {"checked": self._checked}

    def is_checked(self) -> bool:
        """Returns the state of the checkbox. True if checked, False otherwise.

        :returns: Current state of the checkbox.
        :rtype: bool
        """
        return StateJson()[self.widget_id]["checked"]

    def _set(self, checked: bool) -> None:
        """Sets the state of the checkbox.
        This is a private method, so it is not intended to be called directly.
        Use :meth:`check` or :meth:`uncheck` instead.

        :param checked: New state of the checkbox.
        :type checked: bool
        """
        self._checked = checked
        StateJson()[self.widget_id]["checked"] = self._checked
        StateJson().send_changes()

    def check(self) -> None:
        """Sets the state of the checkbox to True."""
        self._set(True)

    def uncheck(self) -> None:
        """Sets the state of the checkbox to False."""
        self._set(False)

    def value_changed(self, func: Callable[[bool], Any]) -> Callable[[], None]:
        """Decorator that handles the event of changing the state of the checkbox.

        :param func: Function that handles the event of changing the state of the checkbox.
        :type func: Callable[[bool], Any]
        :returns: Decorated function.
        :rtype: Callable[[], None]
        """
        route_path = self.get_route_path(Checkbox.Routes.VALUE_CHANGED)
        server = self._sly_app.get_server()
        self._changes_handled = True

        @server.post(route_path)
        def _click():
            res = self.is_checked()
            func(res)

        return _click



from typing import List, Optional, Union

from supervisely import ObjClass, ObjClassCollection
from supervisely.app import DataJson, StateJson
from supervisely.app.widgets import Button, NotificationBox, Widget, generate_id
from supervisely.geometry.alpha_mask import AlphaMask
from supervisely.geometry.any_geometry import AnyGeometry
from supervisely.geometry.bitmap import Bitmap
from supervisely.geometry.closed_surface_mesh import ClosedSurfaceMesh
from supervisely.geometry.cuboid import Cuboid
from supervisely.geometry.cuboid_3d import Cuboid3d
from supervisely.geometry.graph import GraphNodes
from supervisely.geometry.multichannel_bitmap import MultichannelBitmap
from supervisely.geometry.oriented_bbox import OrientedBBox
from supervisely.geometry.point import Point
from supervisely.geometry.point_3d import Point3d
from supervisely.geometry.pointcloud import Pointcloud
from supervisely.geometry.polygon import Polygon
from supervisely.geometry.polyline import Polyline
from supervisely.geometry.rectangle import Rectangle

type_to_shape_text = {
    AnyGeometry: "any shape",
    Rectangle: "rectangle",
    Polygon: "polygon",
    Bitmap: "bitmap (mask)",
    AlphaMask: "alpha mask",
    Polyline: "polyline",
    Point: "point",
    Cuboid: "cuboid",  #
    Cuboid3d: "cuboid 3d",
    Pointcloud: "pointcloud",  #  # "zmdi zmdi-border-clear"
    MultichannelBitmap: "n-channel mask",  # "zmdi zmdi-collection-item"
    Point3d: "point 3d",  # "zmdi zmdi-select-all"
    GraphNodes: "keypoints",
    ClosedSurfaceMesh: "volume (3d mask)",
    OrientedBBox: "oriented bbox",
}


class ClassesMapping(Widget):
    """Widget for mapping source ObjClasses to target ObjClasses with geometry compatibility checks."""

    def __init__(
        self,
        classes: Optional[Union[List[ObjClass], ObjClassCollection]] = [],
        empty_notification: Optional[NotificationBox] = None,
        widget_id: Optional[str] = None,
    ):
        """
        :param classes: List of ObjClass instances or ObjClassCollection.
        :type classes: Optional[Union[List[ObjClass], ObjClassCollection]]
        :param empty_notification: Widget to show when no classes.
        :type empty_notification: NotificationBox, optional
        :param widget_id: Unique widget identifier.
        :type widget_id: str, optional
        """
        if empty_notification is None:
            empty_notification = NotificationBox(
                title="No classes",
                description="No classes to map.",
            )
        self.empty_notification = empty_notification
        self._classes = classes

        self._select_all_btn = Button()
        self._deselect_all_btn = Button()

        self._select_all_btn = Button(
            "Select all",
            button_type="text",
            show_loading=False,
            icon="zmdi zmdi-check-all",
            widget_id=generate_id(),
        )
        self._deselect_all_btn = Button(
            "Deselect all",
            button_type="text",
            show_loading=False,
            icon="zmdi zmdi-square-o",
            widget_id=generate_id(),
        )

        @self._select_all_btn.click
        def _select_all_btn_clicked():
            self.select_all()

        @self._deselect_all_btn.click
        def _deselect_all_btn_clicked():
            self.deselect_all()

        super().__init__(widget_id=widget_id, file_path=__file__)

    def get_json_data(self):
        return {
            "classes": [
                {
                    **cls.to_json(),
                    "shape_text": type_to_shape_text.get(cls.geometry_type).upper(),
                    "default_value": cls.name,
                }
                for cls in self._classes
            ]
        }

    def get_json_state(self):
        return {
            "classes_values": [
                {
                    "value": cls.name,
                    "default": True,
                    "ignore": False,
                    "selected": False,
                }
                for cls in self._classes
            ]
        }

    def set(self, classes):
        self._classes = classes
        self.update_data()
        DataJson().send_changes()
        cur_mapping = self.get_mapping()
        new_mapping_values = []
        for cls in self._classes:
            value = cur_mapping.get(
                cls.name,
                {
                    "value": cls.name,
                    "default": False,
                    "ignore": True,
                    "selected": False,
                },
            )
            new_mapping_values.append(value)
        StateJson()[self.widget_id]["classes_values"] = new_mapping_values
        StateJson().send_changes()

    def get_classes(self):
        return self._classes

    def get_mapping(self):
        classes_values = StateJson()[self.widget_id]["classes_values"]
        if len(classes_values) != len(self._classes):
            self.update_state()
            return self.get_mapping()
        mapping = {
            cls.name: classes_values[idx]
            for idx, cls in enumerate(self._classes)
            if classes_values[idx]["selected"]
        }
        return mapping

    def ignore(self, indexes: List[int]):
        classes_values = StateJson()[self.widget_id]["classes_values"]
        for idx in indexes:
            classes_values[idx] = {"value": "", "default": False, "ignore": True}
        StateJson()[self.widget_id]["classes_values"] = classes_values
        StateJson().send_changes()

    def set_default(self):
        self.update_state()
        StateJson().send_changes()

    def set_mapping(self, mapping: dict):
        cur_mapping = self.get_mapping()
        new_mapping_values = []
        for cls in self._classes:
            cur_value = cur_mapping.get(cls.name, {"value": ""}).get("value")
            new_value = mapping.get(cls.name, cur_value)
            new_mapping_values.append(
                {
                    "value": new_value if new_value != "" else cls.name,
                    "default": new_value == cls.name,
                    "ignore": new_value == "",
                    "selected": new_value != "",
                }
            )
        StateJson()[self.widget_id]["classes_values"] = new_mapping_values
        StateJson().send_changes()

    def select_all(self):
        classes_values = StateJson()[self.widget_id]["classes_values"]
        classes_values: dict
        for value in classes_values:
            value["selected"] = True
        StateJson()[self.widget_id]["classes_values"] = classes_values
        StateJson().send_changes()

    def deselect_all(self):
        classes_values = StateJson()[self.widget_id]["classes_values"]
        classes_values: dict
        for value in classes_values:
            value["selected"] = False
        StateJson()[self.widget_id]["classes_values"] = classes_values
        StateJson().send_changes()

    def select(self, classes):
        classes_values = StateJson()[self.widget_id]["classes_values"]
        classes_values: list
        for idx, cls in enumerate(self._classes):
            classes_values[idx]["selected"] = cls.name in classes
        StateJson()[self.widget_id]["classes_values"] = classes_values
        StateJson().send_changes()



# coding: utf-8
from __future__ import annotations

from typing import Dict, Optional, Union
from uuid import UUID

from supervisely._utils import take_with_default
from supervisely.geometry.geometry import Geometry
from supervisely.pointcloud_annotation.pointcloud_episode_object import (
    PointcloudEpisodeObject,
)
from supervisely.pointcloud_annotation.pointcloud_episode_object_collection import (
    PointcloudEpisodeObjectCollection,
)
from supervisely.pointcloud_annotation.pointcloud_object import PointcloudObject
from supervisely.pointcloud_annotation.pointcloud_object_collection import (
    PointcloudObjectCollection,
)
from supervisely.video_annotation.key_id_map import KeyIdMap
from supervisely.video_annotation.video_figure import VideoFigure


class PointcloudFigure(VideoFigure):
    """Figure in point cloud annotation: 3D geometry (e.g. Cuboid3d) at a frame. Immutable."""

    def __init__(
        self,
        parent_object: Union[PointcloudObject, PointcloudEpisodeObject],
        geometry: Geometry,
        frame_index: Optional[int] = None,
        key: Optional[UUID] = None,
        class_id: Optional[int] = None,
        labeler_login: Optional[str] = None,
        updated_at: Optional[str] = None,
        created_at: Optional[str] = None,
        **kwargs,
    ):
        """
        Figure in point cloud annotation.

        :param parent_object: PointcloudObject or PointcloudEpisodeObject.
        :type parent_object: Union[:class:`~supervisely.pointcloud_annotation.pointcloud_object.PointcloudObject`, :class:`~supervisely.pointcloud_annotation.pointcloud_episode_object.PointcloudEpisodeObject`]
        :param geometry: 3D geometry (Cuboid3d or Pointcloud).
        :type geometry: :class:`~supervisely.geometry.geometry.Geometry`
        :param frame_index: Frame index. Optional for single-frame point clouds.
        :type frame_index: int, optional
        :param key: UUID key. Auto-generated if not provided.
        :type key: uuid.UUID, optional
        :param class_id: Server-side class ID.
        :type class_id: int, optional
        :param labeler_login: Login of user who created the figure.
        :type labeler_login: str, optional
        :param updated_at: Last modification timestamp (ISO format).
        :type updated_at: str, optional
        :param created_at: Creation timestamp (ISO format).
        :type created_at: str, optional

        :Usage Example:

            .. code-block:: python

                import supervisely as sly
                from supervisely.geometry.cuboid_3d import Cuboid3d, Vector3d

                obj_class_car = sly.ObjClass('car', Cuboid3d)
                pointcloud_obj_car = sly.PointcloudObject(obj_class_car)
                cuboid = Cuboid3d(
                    Vector3d(-3.4, 28.9, -0.7),
                    Vector3d(0., 0, -0.03),
                    Vector3d(1.8, 3.9, 1.6),
                )
                figure = sly.PointcloudFigure(pointcloud_obj_car, cuboid, frame_index=10)
                print(figure.to_json())
        """
        super().__init__(
            parent_object,
            geometry,
            frame_index,
            key=key,
            class_id=class_id,
            labeler_login=labeler_login,
            updated_at=updated_at,
            created_at=created_at,
        )
        # @TODO: validate geometry - allowed: only cuboid_3d + point_cloud

    @property
    def parent_object(self) -> Union[PointcloudObject, PointcloudEpisodeObject]:
        """
        PointcloudObject of current PointcloudFigure.

        :returns: Pointcloud object or pointcloud episode object.
        :rtype: :class:`~supervisely.pointcloud_annotation.pointcloud_object.PointcloudObject` or :class:`~supervisely.pointcloud_annotation.pointcloud_episode_object.PointcloudEpisodeObject`

        :Usage Example:

            .. code-block:: python

                pointcloud_obj_car = pointcloud_figure_car.parent_object

                print(pointcloud_obj_car.to_json())
                # Output: {
                #     "key": "d573c6f081544e3da20022d932b259c1",
                #     "classTitle": "car",
                #     "tags": []
                # }
        """

        return super().parent_object

    @property
    def video_object(self) -> None:
        """Not supported for pointcloud."""

        raise NotImplementedError("If you faced this error, please write to technical support.")

    def validate_bounds(self, img_size, _auto_correct=False):
        """Not supported for pointcloud."""

        raise NotImplementedError()

    @classmethod
    def from_json(
        cls,
        data: Dict,
        objects: Union[PointcloudObjectCollection, PointcloudEpisodeObjectCollection],
        frame_index: Optional[int] = None,
        key_id_map: Optional[KeyIdMap] = None,
    ) -> PointcloudFigure:
        """
        Convert a json dict to PointcloudFigure. Read more about `Supervisely format <https://docs.supervisely.com/data-organization/00_ann_format_navi>`_.

        :param data: Dict in json format.
        :type data: dict
        :param objects: Pointcloud object collection or pointcloud episode object collection.
        :type objects: :class:`~supervisely.pointcloud_annotation.pointcloud_object_collection.PointcloudObjectCollection` or :class:`~supervisely.pointcloud_annotation.pointcloud_episode_object_collection.PointcloudEpisodeObjectCollection`
        :param frame_index: Index of Frame to which PointcloudFigure belongs.
        :type frame_index: int
        :param key_id_map: Key ID map.
        :type key_id_map: :class:`~supervisely.video_annotation.key_id_map.KeyIdMap`
        :raises RuntimeError: if point cloudobject ID and pointcloud object key are None, if pointcloud object key and key_id_map are None, if pointcloud object with given id not found in key_id_map
        :returns: Pointcloud figure object.
        :rtype: :class:`~supervisely.pointcloud_annotation.pointcloud_figure.PointcloudFigure`

        :Usage Example:

            .. code-block:: python

                import supervisely as sly
                from supervisely.geometry.cuboid_3d import Cuboid3d, Vector3d

                obj_class_car = sly.ObjClass('car', Cuboid3d)
                pointcloud_obj_car = sly.PointcloudObject(obj_class_car)

                position, rotation, dimension = Vector3d(-3.4, 28.9, -0.7), Vector3d(0., 0, -0.03), Vector3d(1.8, 3.9, 1.6)
                cuboid = Cuboid3d(position, rotation, dimension)
                frame_index = 10
                figure = sly.PointcloudFigure(pointcloud_obj_car, cuboid, frame_index=frame_index)
                pointcloud_figure_json = figure.to_json(save_meta=True)

                new_pointcloud_figure = sly.PointcloudFigure.from_json(
                    pointcloud_figure_json,
                    sly.PointcloudObjectCollection([pointcloud_obj_car]),
                    frame_index
                )
        """

        return super().from_json(data, objects, frame_index, key_id_map)

    def clone(
        self,
        parent_object: Optional[Union[PointcloudObject, PointcloudEpisodeObject]] = None,
        geometry: Optional[Geometry] = None,
        frame_index: Optional[int] = None,
        key: Optional[UUID] = None,
        class_id: Optional[int] = None,
        labeler_login: Optional[str] = None,
        updated_at: Optional[str] = None,
        created_at: Optional[str] = None,
    ) -> PointcloudFigure:
        """
        Makes a copy of PointcloudFigure with new fields, if fields are given, otherwise it will use fields of the original PointcloudFigure.

        :param parent_object: Pointcloud object or pointcloud episode object.
        :type parent_object: :class:`~supervisely.pointcloud_annotation.pointcloud_object.PointcloudObject` or :class:`~supervisely.pointcloud_annotation.pointcloud_episode_object.PointcloudEpisodeObject`, optional
        :param geometry: Label :class:`~supervisely.geometry.geometry.Geometry`.
        :type geometry: :class:`~supervisely.geometry.geometry.Geometry`, optional
        :param frame_index: Index of Frame to which PointcloudFigure belongs.
        :type frame_index: int, optional
        :param key: UUID key associated with the figure.
        :type key: uuid.UUID
        :param class_id: ID of pointcloud object (or pointcloud episode object) to which PointcloudFigure belongs.
        :type class_id: int, optional
        :param labeler_login: Login of the user who created :class:`~supervisely.pointcloud_annotation.pointcloud_figure.PointcloudFigure`.
        :type labeler_login: str, optional
        :param updated_at: Date and Time when PointcloudFigure was modified last. Date Format: Year:Month:Day:Hour:Minute:Seconds. Example: '2021-01-22T19:37:50.158Z'.
        :type updated_at: str, optional
        :param created_at: Date and Time when PointcloudFigure was created. Date Format is the same as in "updated_at" parameter.
        :type created_at: str, optional
        :returns: Pointcloud figure object.
        :rtype: :class:`~supervisely.pointcloud_annotation.pointcloud_figure.PointcloudFigure`

        :Usage Example:

            .. code-block:: python

                import supervisely as sly

                from supervisely.geometry.cuboid_3d import Cuboid3d, Vector3d

                obj_class_car = sly.ObjClass('car', Cuboid3d)
                pointcloud_obj_car = sly.PointcloudObject(obj_class_car)

                position, rotation, dimension = Vector3d(-3.4, 28.9, -0.7), Vector3d(0., 0, -0.03), Vector3d(1.8, 3.9, 1.6)
                cuboid = Cuboid3d(position, rotation, dimension)
                frame_index = 10
                figure = sly.PointcloudFigure(pointcloud_obj_car, cuboid, frame_index=frame_index)

                # Remember that PointcloudFigure object is immutable, and we need to assign new instance of PointcloudFigure to a new variable
                pointcloud_figure_clone = figure.clone(parent_object=pointcloud_obj_car, frame_index=11)

                print(pointcloud_figure_clone.to_json())
                # Output: {
                #     "geometry": {
                #         "dimensions": { "x": 1.8, "y": 3.9, "z": 1.6 },
                #         "position": { "x": -3.4, "y": 28.9, "z": -0.7 },
                #         "rotation": { "x": 0.0, "y": 0, "z": -0.03 }
                #     },
                #     "geometryType": "cuboid_3d",
                #     "key": "4beae1be12624b70ad533c8be7477605",
                #     "objectKey": "c1e1965efc0d4ae9b0b39367b04d637a"
                # }
        """

        return self.__class__(
            parent_object=take_with_default(parent_object, self.parent_object),
            geometry=take_with_default(geometry, self.geometry),
            frame_index=take_with_default(frame_index, self.frame_index),
            key=take_with_default(key, self._key),
            class_id=take_with_default(class_id, self.class_id),
            labeler_login=take_with_default(labeler_login, self.labeler_login),
            updated_at=take_with_default(updated_at, self.updated_at),
            created_at=take_with_default(created_at, self.created_at),
        )



# coding: utf-8


class MetricsBase:
    """Abstract base class for metrics that compare ground-truth and prediction annotations."""

    def add_pair(self, ann_gt, ann_pred):
        raise NotImplementedError()

    def get_metrics(self):
        raise NotImplementedError()

    def get_total_metrics(self):
        raise NotImplementedError()

    def log_total_metrics(self):
        raise NotImplementedError()
